/*
 * Copyright (c) 2003-2020
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: compress.h 3108 2020-05-21 18:07:44Z brachman $
 */

#ifndef _COMPRESS_H_
#define _COMPRESS_H_

#define COMPRESS_DEFAULT_STREAM_INPUT_BUFFER_SIZE		(BUFSIZ * 10)
#define COMPRESS_DEFAULT_STREAM_OUTPUT_BUFFER_SIZE		(BUFSIZ * 10)
#define DECOMPRESS_DEFAULT_STREAM_INPUT_BUFFER_SIZE		(BUFSIZ * 10)
#define DECOMPRESS_DEFAULT_STREAM_OUTPUT_BUFFER_SIZE	(BUFSIZ * 10)

typedef enum Compress_op {
  COMPRESS_PACK   = 1,
  COMPRESS_UNPACK = 2
} Compress_op;

/*
 * COMPRESS_DEFAULT asks for an algorithm to be chosen that is likely to
 * result in the best compression, based on heuristics that can use
 * whichever criteria are at hand (e.g., the size of the input).
 * COMPRESS_BEST asks for the algorithm that achieves the best compression
 * on the input; this may involve compressing the input using
 * the same or different algorithms/parameters.
 *
 * In either case, the algorithm actually selected is set in the
 * operation's configuration (Compress_conf).
 */
typedef enum Compress_alg {
  COMPRESS_ZLIB    = 1,
  COMPRESS_BZ2     = 2,
  COMPRESS_LZMA    = 3,
  COMPRESS_ZSTD    = 4,
  COMPRESS_BROTLI  = 5,
  COMPRESS_DEFAULT = 6,
  COMPRESS_BEST    = 7,
  COMPRESS_NONE    = 0,
  COMPRESS_ERROR   = -1
} Compress_alg;

typedef enum Compress_flag {
  COMPRESS_DISABLED = 0,
  COMPRESS_ENABLED  = 1
} Compress_flag;

/* This is used if COMPRESS_DEFAULT is requested. */
#define COMPRESS_DEFAULT_ALG	COMPRESS_ZLIB

typedef struct Compress_conf Compress_conf;
typedef struct Compress_desc Compress_desc;

typedef void (*dealloc_func)(void *ptr);

/*
 * Compression algorithm description.
 * See compress_tab[] in compress.c.
 */
struct Compress_desc {
  const char *name;		/* The unique, canonical name of the algorithm. */
  const char *aliases;	/* The canonical name plus unique aliases. */
  Compress_alg alg;		/* The unique internal identifier. */
  int flags;
  const char * (*get_version_str)(void);
  int (*compress)(Compress_conf *, void *, size_t, unsigned char **, size_t *);
  void * (*compress_open)(Compress_conf *);
  int (*compress_update)(Compress_conf *, void *, size_t,
						 unsigned char *, size_t, size_t *);
  int (*compress_finish)(Compress_conf *, unsigned char *, size_t, size_t *);
  void (*compress_close)(Compress_conf *);

  int (*decompress)(Compress_conf *, void *, size_t,
					unsigned char **, size_t *);
  void * (*decompress_open)(Compress_conf *);
  int (*decompress_update)(Compress_conf *, void *, size_t,
						   unsigned char *, size_t, size_t *);
  int (*decompress_finish)(Compress_conf *, unsigned char *, size_t, size_t *);
  void (*decompress_close)(Compress_conf *);
};

typedef enum {
  COMPRESS_NO_HEADER   = 0,
  COMPRESS_NEED_HEADER = 1,
  COMPRESS_HAVE_HEADER = 2
} Compress_header;

/*
 * If a call to a compression library function fails, whenever possible
 * error_code will be set to some library-dependent, non-zero value and
 * error_mesg will be set to the corresponding error message string.
 * But some library functions do not return a useful error code, or do so
 * inconsistently, so inspecting error_code to decide if something is wrong
 * is insufficient - rely on this layer's return values.
 *
 * XXX we assume a long int type is sufficiently generic to accomodate all
 * error codes, but if that turns out to be wrong a union might be required.
 */
typedef struct Compress_result {
  ssize_t total_in;			/* Total raw bytes read. */
  ssize_t total_out;		/* Total compressed/decompressed bytes written. */
  /*
   * Amount of memory allocated, in bytes. Explicitly deallocated memory is
   * not deducted, but realloc() memory is.
   */
  size_t total_alloc;		/* Bytes of memory allocated. */
  size_t alloc_count;		/* Number of allocations. */
  size_t dealloc_count;		/* Number of deallocations. */
  size_t realloc_count;		/* Number of reallocations. */
  long error_code;			/* An algorithm-specific error code. */
  const char *error_mesg;	/* If non-NULL, the error message. */

  void *header_start;
  void *data_start;

  /*
   * This is the currently allocated buffer for compression/decompression
   * and its allocated size (not necessarily the size being used).
   * If this structure is returned by compress_pack_close() or
   * compress_unpack_close() the caller is responsible for deallocating
   * buffer if it is not NULL.
   */
  void *buffer;					/* Allocation by custom_alloc/custom_realloc */
  size_t buflen;				/* Allocation size of buffer, in bytes. */
  void (*dealloc)(void *ptr);	/* Its deallocator. */  
} Compress_result;

/*
 * Configuration and context for a compression or decompression operation.
 * If EXPECTED_INPUT_SIZE (EXPECTED_OUTPUT_SIZE) is not zero, it is the
 * caller's expectation of the length in bytes of the input (output) for
 * the compression/decompression function being applied.  This may be used
 * to select buffer sizes, algorithms, etc. to improve efficiency.
 * It is safe to set either or both values to zero.
 * If not zero, it would be best to set them somewhere between their
 * actual value and an approximate upper bound on their actual value.
 * If MALLOC, FREE, or REALLOC are NULL, corresponding default functions
 * are used.
 * Remember that FREE must work with MALLOC and REALLOC.
 */
struct Compress_conf {
  int did_alloc;				/* Non-zero if this needs to be freed. */
  Compress_op op;				/* Compressing or decompressing? */
  Compress_alg alg;				/* Method requested or actually used. */
  Compress_desc *cd;			/* Algorithm description. */

  void *subctx;					/* The compression library's context. */
  int is_finished;				/* Non-zero iff compress/decompress completed. */

  Compress_header header_status;
  size_t header_size;
  size_t header_alloc;

  size_t expected_input_size;	/* */
  size_t expected_output_size;	/* */
  void *(*alloc)(size_t size);	/* Memory allocator to use. */
  void (*dealloc)(void *ptr);	/* And matching deallocator. */
  void *(*realloc)(void *ptr, size_t newsize);	/* And matching reallocator. */
  Compress_result result;		/* Errors and instrumentation. */
};

#define COMPRESS_CTX(CONF)		(CONF->subctx)
#define COMPRESS_BUFFER(CONF)	(CONF->buffer)
#define COMPRESS_BUFLEN(CONF)	(CONF->buflen)

#ifdef __cplusplus
extern "C" {
#endif

extern Compress_desc *compress_lookup_by_alg(Compress_alg alg);
extern Compress_desc *compress_lookup_by_name(const char *name);
extern const char *compress_name(Compress_alg alg);
extern Compress_alg compress_alg(const char *name);
extern Compress_desc *compress_alg_begin(void);
extern Compress_desc *compress_alg_next(Compress_desc *pcd);
extern int compress_alg_is_available(Compress_alg alg);

extern Compress_alg compress_type_from_buffer(unsigned char *buf, size_t buflen,
											  char **type_str);
extern Compress_alg compress_type_from_stream(FILE *, char **type_str);

extern int compress_set_header(Compress_conf *conf, size_t need);
extern int compress_get_header(Compress_conf *conf, void **headerp,
							   size_t *sizep, void **datap);

extern Compress_conf *compress_configure(Compress_conf *conf,
										 Compress_op op, Compress_alg req_alg,
										 void *(*alloc)(size_t size),
										 void (*dealloc)(void *ptr),
										 void *(*realloc)(void *, size_t),
										 size_t approx_inlen,
										 size_t approx_outlen);

extern Compress_result *compress_pack_close(Compress_conf *conf,
											Compress_result *result);
extern Compress_result *compress_unpack_close(Compress_conf *conf,
											  Compress_result *result);

extern const char *compress_version(Compress_alg alg);
extern void compress_info(FILE *fp);
extern char *compress_alg_list(void);

extern void compress_get_default_alloc(void *(**alloc)(size_t size),
									   void (**dealloc)(void *ptr),
									   void *(**realloc)(void *ptr,
														 size_t newsize));
extern void compress_set_default_alloc(void *(*alloc)(size_t size),
									   void (*dealloc)(void *ptr),
									   void *(*realloc)(void *, size_t));
extern void compress_get_conf_alloc(Compress_conf *conf,
									void *(**alloc)(size_t size),
									void (**dealloc)(void *ptr),
									void *(**realloc)(void *ptr,
													  size_t newsize));

extern unsigned char *compress_pack_x(Compress_conf *conf, void *in,
									  size_t inlen,
									  unsigned char **out, size_t *outlen);
extern unsigned char *compress_pack(Compress_alg alg, void *in,
									size_t inlen, unsigned char **out,
									size_t *outlen);
extern Compress_conf *compress_pack_open(Compress_conf *conf);
extern int compress_pack_update(Compress_conf *conf, void *in,
								size_t inlen, unsigned char *out,
								size_t outlen, size_t *buflen);
extern int compress_pack_finish(Compress_conf *conf, unsigned char *out,
								size_t outlen, size_t *buflen);
extern unsigned char *compress_unpack_x(Compress_conf *conf,
										void *in, size_t inlen,
										unsigned char **out, size_t *outlen);
extern unsigned char *compress_unpack(Compress_alg alg, void *in,
									  size_t inlen, unsigned char **out,
									  size_t *outlen);
extern Compress_conf *compress_unpack_open(Compress_conf *conf, FILE *);
extern int compress_unpack_update(Compress_conf *conf, void *in,
								  size_t inlen, unsigned char *out,
								  size_t outlen, size_t *buflen);
extern int compress_unpack_finish(Compress_conf *conf, unsigned char *out,
								  size_t outlen, size_t *buflen);

extern ssize_t compress_fpack(FILE *fp_in, FILE *fp_out, Compress_alg alg);
extern ssize_t compress_fpack_x(Compress_conf *conf,
								FILE *fp_in, unsigned char *, size_t,
								FILE *fp_out, unsigned char *, size_t);

extern ssize_t compress_funpack(FILE *fp_in, FILE *fp_out, Compress_alg alg);
extern ssize_t compress_funpack_x(Compress_conf *conf,
								  FILE *fp_in, unsigned char *, size_t,
								  FILE *fp_out, unsigned char *, size_t);

#ifdef __cplusplus
}
#endif

#endif
