/*
 * Note: this should be standalone - do not bring in any DACS dependencies.
 */

/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 *
 * See crypto_sha512.c for applicable open source licensing.
 * $Id: crypto_sha512.h 3294 2024-03-21 21:16:25Z brachman $
 */

#ifndef _CRYPTO_SHA512_H_
#define _CRYPTO_SHA512_H_

#include <stdint.h>

typedef enum {
  SHA512_224_alg     = 0,
  SHA512_256_alg     = 1,
  SHA512_384_alg     = 2,
  SHA512_512_alg     = 3,
  SHA512_512t_alg    = 4,
  SHA512_INVALID_alg = 5
} SHA512_alg;

enum {
  SHA512_224_DIGEST_BYTE_LENGTH = 28,
  SHA512_256_DIGEST_BYTE_LENGTH = 32,
  SHA512_384_DIGEST_BYTE_LENGTH = 48,
  SHA512_512_DIGEST_BYTE_LENGTH = 64,
  SHA512_224_DIGEST_BIT_LENGTH  = (SHA512_224_DIGEST_BYTE_LENGTH * 8),
  SHA512_256_DIGEST_BIT_LENGTH  = (SHA512_256_DIGEST_BYTE_LENGTH * 8),
  SHA512_384_DIGEST_BIT_LENGTH  = (SHA512_384_DIGEST_BYTE_LENGTH * 8),
  SHA512_512_DIGEST_BIT_LENGTH  = (SHA512_512_DIGEST_BYTE_LENGTH * 8),
  SHA512_DIGEST_BYTE_LENGTH     = SHA512_512_DIGEST_BYTE_LENGTH,
  SHA512_DIGEST_BIT_LENGTH      = (SHA512_DIGEST_BYTE_LENGTH * 8)
};

typedef struct SHA512_ctx {
  SHA512_alg alg;
  unsigned int digest_bits;				/* This is a t-bit hash. */
  uint64_t sz[2];
  uint64_t counter[8];
  unsigned char save[128];
} SHA512_ctx;

#ifdef __cplusplus
extern "C" {
#endif

extern int sha512_ctx_digest_length(SHA512_ctx *m);
extern int sha512_digest_length(SHA512_alg alg, unsigned int t);

extern int sha512_init_alg(SHA512_ctx *, SHA512_alg);
extern void sha512_ctx_zap(SHA512_ctx *m);

extern void sha512_224_init(SHA512_ctx *m);
extern void sha512_256_init(SHA512_ctx *m);
extern void sha512_384_init(SHA512_ctx *m);
extern void sha512_512_init(SHA512_ctx *m);
extern void sha512_init(SHA512_ctx *m);		/* Synonym for sha512_512_init(). */
extern int sha512t_init(SHA512_ctx *m, unsigned int t);

extern unsigned char *sha512_224(const void *input, size_t input_len,
								 void *result);
extern unsigned char *sha512_256(const void *input, size_t input_len,
								 void *result);
extern unsigned char *sha512_384(const void *input, size_t input_len,
								 void *result);
extern unsigned char *sha512_t(unsigned int t, const void *input,
							   size_t input_len, void *result);
extern unsigned char *sha512(const void *input, size_t input_len, void *result);

extern unsigned char *sha512_224_final(SHA512_ctx *, void *);
extern unsigned char *sha512_256_final(SHA512_ctx *, void *);
extern unsigned char *sha512_384_final(SHA512_ctx *, void *);
extern unsigned char *sha512_512_final(SHA512_ctx *, void *);

extern void sha512_update(SHA512_ctx *, const void *, size_t);
extern unsigned char *sha512_final(SHA512_ctx *, void *);

extern int sha512_t_result(unsigned char *res, unsigned int t, int *nbytes,
						   int *nbits, int *bitval);

extern int do_crypto_sha512_tests(void);

#ifdef __cplusplus
}
#endif

#endif
