/*
 * Copyright (c) 2003-2023
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: dacs_glob.h 3244 2023-06-28 22:52:53Z brachman $
 */

/*
 * This is a slightly enhanced replacement for the BSD implementation.
 * XXX cannot co-exist with glob.h and the system implementation because
 * the same names/symbols are used; this may be addressed in a future version
 */

/*-
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Copyright (c) 1989, 1993
 *  The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Guido van Rossum.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *  @(#)glob.h  8.1 (Berkeley) 6/2/93
 * $FreeBSD$
 */

#ifndef _GLOB_H_
#define _GLOB_H_

#include <sys/cdefs.h>

#ifdef __FreeBSD__
#include <sys/_types.h>
#else
/* For non-FreeBSD systems... */
#ifndef __BSD_VISIBLE
#define __BSD_VISIBLE   1
#endif

#ifndef __POSIX_VISIBLE
#define __POSIX_VISIBLE 199209
#endif

#if defined(__linux__)
#include <linux/limits.h>
#elif defined(__APPLE__)
#include <sys/types.h>
typedef u_int64_t __size_t;
#define _SIZE_T_DECLARED
#endif

#ifndef _SIZE_T_DECLARED
typedef __size_t    size_t;
#define _SIZE_T_DECLARED
#endif
#endif

struct stat;

typedef struct glob_t glob_t;
struct glob_t {
  size_t gl_pathc;    /* Count of total paths so far. */
  size_t gl_matchc;   /* Count of paths matching pattern. */
  size_t gl_offs;     /* Reserved at beginning of gl_pathv. */
  int gl_flags;       /* Copy of flags parameter to glob. */
  char **gl_pathv;    /* List of paths matching pattern. */
  /* Copy of errfunc parameter to glob. */
  int (*gl_errfunc)(const char *, int);

  /*
   * Alternate filesystem access methods for glob; caller-provided replacement
   * versions of closedir(3), readdir(3), opendir(3), stat(2), and lstat(2).
   * These must all be specified if this capability is enabled through the
   * GLOB_ALTDIRFUNC flag to glob() and globmatch().
   *
   * The original code has been extended to pass a pointer to the glob_t
   * structure, to which the gl_user_ptr element has been added.
   * This allows arbitrary data to be shared amongst these functions.
   */
  void (*gl_closedir)(glob_t *, void *);
  struct dirent *(*gl_readdir)(glob_t *, void *);
  void *(*gl_opendir)(glob_t *, const char *);
  int (*gl_lstat)(glob_t *, const char *, struct stat *);
  int (*gl_stat)(glob_t *, const char *, struct stat *);

  /*
   * In some runtime contexts, when tilde expansion is needed for
   * the "implied user" (i.e., "~/foo" versus an explicitly named
   * user such as "~bobo/foo"), the home directory path must be provided
   * since it may not be available or desirable to put in the environment.
   *
   * If the GLOB_TILDE_HOME flag is not set, the normal algorithm is used.
   * If GLOB_TILDE_HOME is set:
   *   if GL_TILDE_HOME is NULL, then no substitution is applied to the
   *   "~/..." construct and the result is the same as when the normal
   *   algorithm cannot do the substitution;
   *   if GL_TILDE_HOME is non-NULL, it replaces the tilde.
   */
  char *gl_tilde_home;

  /* Private, for passing arbitrary data to alternate access methods. */
  void *gl_user_ptr;
};

#if __POSIX_VISIBLE >= 199209
/* Believed to have been introduced in 1003.2-1992 */
enum {
  GLOB_APPEND   = 0x0001,  /* Append to output from previous call. */
  GLOB_DOOFFS   = 0x0002,  /* Use gl_offs. */
  GLOB_ERR      = 0x0004,  /* Return on error. */
  GLOB_MARK     = 0x0008,  /* Append / to matching directories. */
  GLOB_NOCHECK  = 0x0010,  /* Return pattern itself if nothing matches. */
  GLOB_NOSORT   = 0x0020,  /* Don't sort. */
  GLOB_NOESCAPE = 0x2000   /* Disable backslash escaping. */
};

/* Error values returned by glob() and globmatch(). */
enum {
  GLOB_NOSPACE = -1,    /* Malloc call failed. */
  GLOB_ABORTED = -2,    /* Unignored error. */
  GLOB_NOMATCH = -3,    /* No match and GLOB_NOCHECK was not set. */
  GLOB_NOSYS   = -4     /* Obsolete: source comptability only. */
};
#endif /* __POSIX_VISIBLE >= 199209 */

#if __BSD_VISIBLE
enum {
  GLOB_ALTDIRFUNC = 0x0040,  /* Use alternately specified directory funcs. */
  GLOB_BRACE      = 0x0080,  /* Expand braces ala csh. */
  GLOB_MAGCHAR    = 0x0100,  /* Pattern had globbing characters. */
  GLOB_NOMAGIC    = 0x0200,  /* GLOB_NOCHECK without magic chars (csh). */
  GLOB_QUOTE      = 0x0400,  /* Quote special chars with \. */
  GLOB_TILDE      = 0x0800,  /* Expand tilde names from the passwd file. */
  GLOB_LIMIT      = 0x1000   /* limit number of returned paths */
};

/* Source compatibility, these are the old names. */
enum {
  GLOB_MAXPATH = GLOB_LIMIT,
  GLOB_ABEND   = GLOB_ABORTED
};

/* For the DACS version. */
enum {
  GLOB_TILDE_HOME = 0x4000   /* Always honour the gl_tilde_home element. */
};

#endif /* __BSD_VISIBLE */

extern int globmatch(const char * __restrict str,
					 const char * __restrict pattern, int flags);
extern int glob(const char * __restrict, int,
				int (*)(const char *, int), glob_t * __restrict);
extern glob_t *globinit(glob_t *, int);
extern void globfree(glob_t *);

#endif
