/*
 * Copyright (c) 2003-2025
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * This is the very start of work on OAUTH authentication that did
 * not go anywhere. It is unlikely to ever be completed since
 * OpenID Connect authentication is being developed.
 * This work is retained for hysterical reasons.
 */

#include "auth.h"

#ifndef lint
static const MAYBE_UNUSED char copyright[] =
"Copyright (c) 2003-2025\n\
Distributed Systems Software.  All rights reserved.";
static const MAYBE_UNUSED char revid[] =
  "$Id: auth_oauth.c 3386 2025-04-01 00:30:08Z brachman $";
#endif

static char *log_module_name = "auth_oauth";

/*
 * Also see:
 * RFC 6749 https://datatracker.ietf.org/doc/html/rfc6749
 * RFC 6750 https://datatracker.ietf.org/doc/html/rfc6750
 * RFC 8252 https://datatracker.ietf.org/doc/html/rfc8252
 *
 * JSON Web Token (JWT)
 * RFC 7159 https://datatracker.ietf.org/doc/html/rfc7159
 * RFC 7515 https://datatracker.ietf.org/doc/html/rfc7515
 * RFC 7519 https://datatracker.ietf.org/doc/html/rfc7519
 * RFC 7523 https://datatracker.ietf.org/doc/html/rfc7523
 * https://openid.net/specs/openid-connect-core-1_0.html#IDToken
 */

static int debug_oauth = 0;

static char *
oauth_encode(char *str)
{
  char *enc_str;
  static char *oauth_chars =
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-._~";

  enc_str = percent_encode_other_chars(str, oauth_chars, 0);

  return(enc_str);
}

static int
oauth_param_compar(const void *ap, const void *bp)
{
  int st;
  Uri_query_arg *a, *b;

  a = *(Uri_query_arg **) ap;
  b = *(Uri_query_arg **) bp;

  if ((st = strcmp(a->name, b->name)) == 0)
	st = strcmp(a->value, b->value);

  return(st);
}

static char *
oauth_normalize_params(Dsvec *dsv_args, Dsvec *dsv_normalized_args)
{
  int i;
  Ds ds;
  Uri_query_arg *arg, *enc_arg;

  for (i = 0; i < dsvec_len(dsv_args); i++) {
	arg = (Uri_query_arg *) dsvec_ptr_index(dsv_args, i);
	enc_arg = ALLOC(Uri_query_arg);
	enc_arg->name = oauth_encode(arg->name);
	enc_arg->value = oauth_encode(arg->value);
	dsvec_add_ptr(dsv_normalized_args, enc_arg);
	if (debug_oauth > 1) {
	  fprintf(stderr,
			  "RawArg%d: %s=\"%s\"\n", i, arg->name, arg->value);
	  fprintf(stderr,
			  "EncArg%d: %s=\"%s\"\n", i, enc_arg->name, enc_arg->value);
	}
  }

  dsvec_sort(dsv_normalized_args, oauth_param_compar);

  ds_init(&ds);
  for (i = 0; i < dsvec_len(dsv_normalized_args); i++) {
	char *arg_str;

	arg = (Uri_query_arg *) dsvec_ptr_index(dsv_normalized_args, i);
	arg_str = ds_xprintf("%s=%s",
						 arg->name, (arg->value == NULL) ? "" : arg->value);
	ds_asprintf(&ds, "%s%s", (i == 0) ? "" : "&", arg_str);
  }

  return(ds_buf(&ds));
}

typedef enum {
  OAUTH_HMAC_SHA1 = 0,
  OAUTH_RSA_SHA1  = 1,
  OAUTH_PLAINTEXT = 2
} Oauth_sig_method;

/*
 * RFC 5849 S3.4
 *
 * XXX
 * http://hueniverse.com/2012/07/oauth-2-0-and-the-road-to-hell/
 * https://tech.slashdot.org/story/12/07/28/1538212/oauth-20-standard-editor-quits-takes-name-off-spec
 * 
 */
static char *
auth_oauth_signature(char *method, Uri *uri, Dsvec *oauth_args,
					 char *body_arg_str, Dsvec *normalized_oauth_args,
					 Oauth_sig_method sig_method, char *client_shared_secret,
					 char *token_shared_secret, char *private_key_file,
					 char *passphrase)
{
  unsigned int outp_len;
  char *base_string_uri, *enc, *key;
  char *normalized_param_string, *signature_base_string;
  unsigned char *outp;
  Ds ds;
  Dsvec *args, *body_args;

  ds_init(&ds);
  ds_asprintf(&ds, "%s://%s", uri->scheme, uri->host);
  if (uri->port_given != NULL) {
	if (uri->is_http) {
	  if (uri->port != HTTP_DEFAULT_PORT)
		ds_asprintf(&ds, ":%s", uri->port_given);
	}
	else if (uri->is_https) {
	  if (uri->port != HTTP_DEFAULT_SSL_PORT)
		ds_asprintf(&ds, ":%s", uri->port_given);
	}
	else
	  ds_asprintf(&ds, ":%s", uri->port_given);
  }
  if (uri->path == NULL || *uri->path == '\0')
	ds_asprintf(&ds, "/");
  else
	ds_asprintf(&ds, "%s", uri->path);

  base_string_uri = ds_buf(&ds);
  if (debug_oauth)
	fprintf(stderr, "base_string_uri=\"%s\"\n", base_string_uri);

  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  dsvec_append(args, uri->query_args);
  dsvec_append(args, oauth_args);
  if (body_arg_str != NULL) {
	if ((body_args = parse_query_string(strdup(body_arg_str))) == NULL)
	  return(NULL);
	dsvec_append(args, body_args);
  }

  normalized_param_string = oauth_normalize_params(args, normalized_oauth_args);
  if (debug_oauth)
	fprintf(stderr, "normalized_params=\"%s\"\n", normalized_param_string);

  ds_reset_buf(&ds);
  ds_asprintf(&ds, "%s&%s&%s",
			  oauth_encode(strtoupper(method)),
			  oauth_encode(base_string_uri),
			  oauth_encode(normalized_param_string));
			  

  signature_base_string = ds_buf(&ds);
  if (debug_oauth)
	fprintf(stderr, "signature_base_string=\"%s\"\n", signature_base_string);

  if (sig_method == OAUTH_HMAC_SHA1) {
	key = ds_xprintf("%s&%s",
					 oauth_encode(client_shared_secret),
					 oauth_encode(token_shared_secret));
	outp = crypto_hmac("SHA1", (unsigned char *) key, strlen(key),
					   ds_ucbuf(&ds), ds_len(&ds) - 1,
					   NULL, &outp_len);
	mime_encode_base64(outp, outp_len, &enc);
	if (debug_oauth)
	  fprintf(stderr, "HMAC-SHA1(\"%s\", \"%s\") = \"%s\"\n",
			  key, signature_base_string, enc);
  }
  else if (sig_method == OAUTH_RSA_SHA1) {
	Ds *sig;
	RSA *priv_key;

	priv_key = pem_load_private_key(private_key_file, NULL);
	if (priv_key == NULL)
	  return(NULL);
	if ((sig = crypto_sign_buf(&ds, priv_key)) == NULL)
	  return(NULL);
	mime_encode_base64(ds_ucbuf(sig), ds_len(sig), &enc);
	if (debug_oauth)
	  fprintf(stderr, "RSA-SHA1(\"%s\", \"%s\") = \"%s\"\n",
			  private_key_file, signature_base_string, enc);
  }
  else {
	enc = ds_xprintf("%s&%s",
					 oauth_encode(client_shared_secret),
					 oauth_encode(token_shared_secret));
	if (debug_oauth)
	  fprintf(stderr, "PLAINTEXT(\"%s\") = \"%s\"\n",
			  signature_base_string, enc);
  }

  return(enc);
}

/*
 * Also see:
 * http://oauth.googlecode.com/svn/code/javascript/example/signature.html
 */
int
auth_oauth_do_oauth_test(void)
{
  int i, testnum;
  unsigned int outp_len;
  char *client_shared_secret, *enc, *signature, *token_shared_secret;
  char *oauth_signature;
  unsigned char *outp;
  Dsvec *args, *normalized_args;
  Uri *uri;
  Uri_query_arg *arg;

  testnum = 0;

  /*
   * Tests from: https://wiki.oauth.net/TestCases
   */
  /* Test S3.1 and 3.4.1.1 */
  fprintf(stderr, "Test %d (HMAC-SHA1)... ", ++testnum);
  outp = crypto_hmac("SHA1", (unsigned char *) "cs&", 3,
					 (unsigned char *) "bs", 2, NULL, &outp_len);
  mime_encode_base64(outp, outp_len, &enc);
  if (!streq(enc, "egQqG5AJep5sJ7anhXju1unge2I=")) {
	fprintf(stderr, "failed\n");
	return(-1);
  }
  fprintf(stderr, "ok\n");

  fprintf(stderr, "Test %d (HMAC-SHA1)... ", ++testnum);
  outp = crypto_hmac("SHA1", (unsigned char *) "cs&ts", 5,
					 (unsigned char *) "bs", 2, NULL, &outp_len);
  mime_encode_base64(outp, outp_len, &enc);
  if (!streq(enc, "VZVjXceV7JgPq/dOTnNmEfO0Fv8=")) {
	fprintf(stderr, "failed\n");
	return(-1);
  }
  fprintf(stderr, "ok\n");

  fprintf(stderr, "Test %d (vacation.jpg)... ", ++testnum);
  client_shared_secret = "kd94hf93k423kf44";
  token_shared_secret = "pfkkdhi9sl3r4s00";
  uri = uri_parse("http://photos.example.net/photos");
  uri_add_query_arg(uri->query_args, "file", "vacation.jpg");
  uri_add_query_arg(uri->query_args, "size", "original");

  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  uri_add_query_arg(args, "oauth_consumer_key", "dpf43f3p2l4k3l03");
  uri_add_query_arg(args, "oauth_token", "nnch734d00sl2jdk");
  uri_add_query_arg(args, "oauth_signature_method", "HMAC-SHA1");
  uri_add_query_arg(args, "oauth_timestamp", "1191242096");
  uri_add_query_arg(args, "oauth_nonce", "kllo9940pd9333jh");
  uri_add_query_arg(args, "oauth_version", "1.0");

  normalized_args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  signature = auth_oauth_signature("GET", uri, args,
								   NULL,
								   normalized_args, OAUTH_HMAC_SHA1,
								   client_shared_secret, token_shared_secret,
								   NULL, NULL);
  if (debug_oauth) {
	fprintf(stderr, "Authorization: OAuth\n");
	for (i = 0; i < dsvec_len(normalized_args); i++) {
	  arg = dsvec_ptr_index(normalized_args, i);
	  fprintf(stderr, "  %s=\"%s\",\n", arg->name, arg->value);
	}
  }
  if (signature == NULL) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  if (!streq(signature, "tR3+Ty81lMeYAr/Fid0kMTYa/WM=")) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  fprintf(stderr, "ok\n");

  fprintf(stderr, "Test %d (vacaction.jpg)... ", ++testnum);
  client_shared_secret = "";
  token_shared_secret = "";
  uri = uri_parse("http://photos.example.net/photos");
  uri_add_query_arg(uri->query_args, "file", "vacaction.jpg");
  uri_add_query_arg(uri->query_args, "size", "original");

  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  uri_add_query_arg(args, "oauth_consumer_key", "dpf43f3p2l4k3l03");
  uri_add_query_arg(args, "oauth_signature_method", "RSA-SHA1");
  uri_add_query_arg(args, "oauth_timestamp", "1196666512");
  uri_add_query_arg(args, "oauth_nonce", "13917289812797014437");
  uri_add_query_arg(args, "oauth_version", "1.0");

  normalized_args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  signature = auth_oauth_signature("GET", uri, args,
								   NULL,
								   normalized_args, OAUTH_RSA_SHA1,
								   client_shared_secret, token_shared_secret,
								   "tests/oauth_test_private_key.pem", NULL);
  if (debug_oauth) {
	fprintf(stderr, "Authorization: OAuth\n");
	for (i = 0; i < dsvec_len(normalized_args); i++) {
	  arg = dsvec_ptr_index(normalized_args, i);
	  fprintf(stderr, "  %s=\"%s\",\n", arg->name, arg->value);
	}
  }
  if (signature == NULL) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  if (!streq(signature, "jvTp/wX1TYtByB1m+Pbyo0lnCOLIsyGCH7wke8AUs3BpnwZJtAuEJkvQL2/9n4s5wUmUl4aCI4BwpraNx4RtEXMe5qg5T1LVTGliMRpKasKsW//e+RinhejgCuzoH26dyF8iY2ZZ/5D1ilgeijhV/vBka5twt399mXwaYdCwFYE=")) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  fprintf(stderr, "ok\n");

  /* RFC 5849 test - S2.1 */
  fprintf(stderr, "Test %d (S2.1)... ", ++testnum);
  client_shared_secret = "ja893SD9";
  token_shared_secret = "";

  uri = uri_parse("https://server.example.com/request_temp_credentials");

  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  uri_add_query_arg(args, "oauth_consumer_key", "jd83jd92dhsh93js");
  uri_add_query_arg(args, "oauth_signature_method", "PLAINTEXT");
  uri_add_query_arg(args, "oauth_callback", "http://client.example.net/cb?x=1");

  normalized_args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  signature = auth_oauth_signature("POST", uri, args,
								   NULL,
								   normalized_args, OAUTH_PLAINTEXT,
								   client_shared_secret, token_shared_secret,
								   NULL, NULL);
  oauth_signature = oauth_encode(signature);
  if (debug_oauth) {
	fprintf(stderr, "Authorization: OAuth\n");
	for (i = 0; i < dsvec_len(normalized_args); i++) {
	  arg = dsvec_ptr_index(normalized_args, i);
	  fprintf(stderr, "  %s=\"%s\",\n", arg->name, arg->value);
	}
	fprintf(stderr, "  oauth_signature=\"%s\"\n", oauth_signature);
  }
  if (!streq(oauth_signature, "ja893SD9%26")) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  fprintf(stderr, "ok\n");

  /* RFC 5849 test - S2.3 */
  fprintf(stderr, "Test %d (S2.3)... ", ++testnum);
  client_shared_secret = "ja893SD9";
  token_shared_secret = "xyz4992k83j47x0b";

  uri = uri_parse("https://server.example.com/request_token");

  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  uri_add_query_arg(args, "oauth_consumer_key", "jd83jd92dhsh93js");
  uri_add_query_arg(args, "oauth_token", "hdk48Djdsa");
  uri_add_query_arg(args, "oauth_signature_method", "PLAINTEXT");
  uri_add_query_arg(args, "oauth_verifier", "473f82d3");

  normalized_args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  signature = auth_oauth_signature("POST", uri, args,
								   NULL,
								   normalized_args, OAUTH_PLAINTEXT,
								   client_shared_secret, token_shared_secret,
								   NULL, NULL);
  oauth_signature = oauth_encode(signature);
  if (debug_oauth) {
	fprintf(stderr, "Authorization: OAuth\n");
	for (i = 0; i < dsvec_len(normalized_args); i++) {
	  arg = dsvec_ptr_index(normalized_args, i);
	  fprintf(stderr, "  %s=\"%s\",\n", arg->name, arg->value);
	}
	fprintf(stderr, "  oauth_signature=\"%s\"\n", oauth_signature);
  }
  if (!streq(oauth_signature, "ja893SD9%26xyz4992k83j47x0b")) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  fprintf(stderr, "ok\n");

  /* RFC 5849 test - S3.1 and 3.4.1.1 */
  fprintf(stderr, "Test %d (S3.1/3.4.1.1)... ", ++testnum);
  client_shared_secret = "j49sk3j29djd";
  token_shared_secret = "dh893hdasih9";

  uri = uri_parse("http://example.com/request");
  uri_add_query_arg(uri->query_args, "b5", "=%3D");
  uri_add_query_arg(uri->query_args, "a3", "a");
  uri_add_query_arg(uri->query_args, "c@", "");
  uri_add_query_arg(uri->query_args, "a2", "r b");

  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  uri_add_query_arg(args, "oauth_consumer_key", "9djdj82h48djs9d2");
  uri_add_query_arg(args, "oauth_token", "kkk9d7dh3k39sjv7");
  uri_add_query_arg(args, "oauth_signature_method", "HMAC-SHA1");
  uri_add_query_arg(args, "oauth_timestamp", "137131201");
  uri_add_query_arg(args, "oauth_nonce", "7d8f3e4a");

  normalized_args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  signature = auth_oauth_signature("POST", uri, args,
								   "c2&a3=2 q",
								   normalized_args, OAUTH_HMAC_SHA1,
								   client_shared_secret, token_shared_secret,
								   NULL, NULL);
  oauth_signature = oauth_encode(signature);
  if (debug_oauth) {
	fprintf(stderr, "Authorization: OAuth\n");
	for (i = 0; i < dsvec_len(normalized_args); i++) {
	  arg = dsvec_ptr_index(normalized_args, i);
	  fprintf(stderr, "  %s=\"%s\",\n", arg->name, arg->value);
	}
	fprintf(stderr, "  oauth_signature=\"%s\"\n", oauth_signature);
  }
  if (!streq(oauth_signature, "r6%2FTJjbCOr97%2F%2BUU0NsvSne7s5g%3D")) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  fprintf(stderr, "ok\n");

  /* RFC 5849 test - S1.2 */
  fprintf(stderr, "Test %d (S1.2-1)... ", ++testnum);
  client_shared_secret = "kd94hf93k423kf44";
  token_shared_secret = "";

  uri = uri_parse("https://photos.example.net/initiate");
  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  uri_add_query_arg(args, "oauth_consumer_key", "dpf43f3p2l4k3l03");
  uri_add_query_arg(args, "oauth_signature_method", "HMAC-SHA1");
  uri_add_query_arg(args, "oauth_timestamp", "137131200");
  uri_add_query_arg(args, "oauth_nonce", "wIjqoS");
  uri_add_query_arg(args, "oauth_callback", "http://printer.example.com/ready");

  normalized_args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  signature = auth_oauth_signature("post", uri, args, NULL,
								   normalized_args, OAUTH_HMAC_SHA1,
								   client_shared_secret, token_shared_secret,
								   NULL, NULL);
  oauth_signature = oauth_encode(signature);
  if (debug_oauth) {
	fprintf(stderr, "Authorization: OAuth\n");
	for (i = 0; i < dsvec_len(normalized_args); i++) {
	  arg = dsvec_ptr_index(normalized_args, i);
	  fprintf(stderr, "  %s=\"%s\",\n", arg->name, arg->value);
	}
	fprintf(stderr, "  oauth_signature=\"%s\"\n", oauth_signature);
  }
  if (!streq(oauth_signature, "74KNZJeDHnMBp0EMJ9ZHt%2FXKycU%3D")) {
	fprintf(stderr, "failed\n");
	return(-1);
  }
  fprintf(stderr, "ok\n");

  /* RFC 5849 test -  S1.2 */
  fprintf(stderr, "Test %d (S1.2-2)... ", ++testnum);
  client_shared_secret = "kd94hf93k423kf44";
  token_shared_secret = "hdhd0244k9j7ao03";
  uri = uri_parse("https://photos.example.net/token");
  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  uri_add_query_arg(args, "oauth_consumer_key", "dpf43f3p2l4k3l03");
  uri_add_query_arg(args, "oauth_token", "hh5s93j4hdidpola");
  uri_add_query_arg(args, "oauth_signature_method", "HMAC-SHA1");
  uri_add_query_arg(args, "oauth_timestamp", "137131201");
  uri_add_query_arg(args, "oauth_nonce", "walatlh");
  uri_add_query_arg(args, "oauth_verifier", "hfdp7dh39dks9884");

  normalized_args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  signature = auth_oauth_signature("post", uri, args, NULL,
								   normalized_args, OAUTH_HMAC_SHA1,
								   client_shared_secret, token_shared_secret,
								   NULL, NULL);
  oauth_signature = oauth_encode(signature);
  if (debug_oauth) {
	fprintf(stderr, "Authorization: OAuth\n");
	for (i = 0; i < dsvec_len(normalized_args); i++) {
	  arg = dsvec_ptr_index(normalized_args, i);
	  fprintf(stderr, "  %s=\"%s\",\n", arg->name, arg->value);
	}
	fprintf(stderr, "  oauth_signature=\"%s\"\n", oauth_signature);
  }
  if (!streq(oauth_signature, "gKgrFCywp7rO0OXSjdot%2FIHF7IU%3D")) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  fprintf(stderr, "ok\n");

  /* RFC 5849 test -  S1.2 */
  fprintf(stderr, "Test %d (S1.2-3)... ", ++testnum);
  client_shared_secret = "kd94hf93k423kf44";
  token_shared_secret = "pfkkdhi9sl3r4s00";
  uri = uri_parse("http://photos.example.net/photos");
  uri_add_query_arg(uri->query_args, "file", "vacation.jpg");
  uri_add_query_arg(uri->query_args, "size", "original");
  args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  uri_add_query_arg(args, "oauth_consumer_key", "dpf43f3p2l4k3l03");
  uri_add_query_arg(args, "oauth_token", "nnch734d00sl2jdk");
  uri_add_query_arg(args, "oauth_signature_method", "HMAC-SHA1");
  uri_add_query_arg(args, "oauth_timestamp", "137131202");
  uri_add_query_arg(args, "oauth_nonce", "chapoH");

  normalized_args = dsvec_init(NULL, sizeof(Uri_query_arg *));
  signature = auth_oauth_signature("get", uri, args, NULL,
								   normalized_args, OAUTH_HMAC_SHA1,
								   client_shared_secret, token_shared_secret,
								   NULL, NULL);
  oauth_signature = oauth_encode(signature);
  if (debug_oauth) {
	fprintf(stderr, "Authorization: OAuth\n");
	for (i = 0; i < dsvec_len(normalized_args); i++) {
	  arg = dsvec_ptr_index(normalized_args, i);
	  fprintf(stderr, "  %s=\"%s\",\n", arg->name, arg->value);
	}
	fprintf(stderr, "  oauth_signature=\"%s\"\n", oauth_signature);
  }
  if (!streq(oauth_signature, "MdpQcU8iPSUjWoN%2FUDMsK2sui9I%3D")) {
	fprintf(stderr, "failed\n");
	fprintf(stderr, "client_shared_secret=\"%s\", token_shared_secret=\"%s\"\n",
			client_shared_secret, token_shared_secret);
	return(-1);
  }
  fprintf(stderr, "ok\n");

  return(0);
}
