
/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * XXX Functionality to support authentication via the
 * OpenID Connect Relying Party protocol will be put here.
 * It is not implemented in this version of DACS.
 */

#include "auth.h"

#ifndef lint
static const MAYBE_UNUSED char copyright[] =
"Copyright (c) 2003-2024\n\
Distributed Systems Software.  All rights reserved.";
static const MAYBE_UNUSED char revid[] =
  "$Id: auth_oidc.c 3342 2024-09-23 20:31:41Z brachman $";
#endif

static char *log_module_name = "auth_oidc";

static MAYBE_UNUSED void
dump_ub(void *ptr, int n)
{
  int i;
  unsigned char *b;

  b = (unsigned char *) ptr;
  printf("[");
  for (i = 0; i < n; i++) {
	printf("%3u", b[i]);
	if (i == (n - 1))
	  printf("]\n");
	else if (((i + 1) % 14) == 0)
	  printf(",\n ");
	else
	  printf(", ");
  }

}

static MAYBE_UNUSED void
dump_hb(void *ptr, int n)
{
  int i;
  unsigned char *b;

  b = (unsigned char *) ptr;
  printf("[");
  for (i = 0; i < n; i++) {
	printf("%02x", b[i]);
	if (i == (n - 1))
	  printf("]\n");
	else if (((i + 1) % 14) == 0)
	  printf(",\n ");
	else
	  printf(", ");
  }

}

static MAYBE_UNUSED void
dump_c(const void *ptr, int n)
{
  int i;
  unsigned char *b;

  b = (unsigned char *) ptr;
  printf("[");
  for (i = 0; i < n; i++) {
	printf("%c", b[i]);
	if (i == (n - 1))
	  printf("]\n");
	else if (((i + 1) % 14) == 0)
	  printf(",\n ");
	else
	  printf(", ");
  }

}

/*
 * RFC 7515 A.1.1
 *
 * Also see:
 *  https://www.cryptopp.com/wiki/JSON_Web_Signature
 */
static const char *jws_protected_header =
  "{\"typ\":\"JWT\",\x0d\x0a \"alg\":\"HS256\"}";

static const char *jws_payload =
  "{\"iss\":\"joe\",\x0d\x0a \"exp\":1300819380,\x0d\x0a \"http://example.com/is_root\":true}";

static const char *jwk_key_oct =
  "AyM1SysPpbyDfgZld3umj1qzKObwVMkoqQ-EstJQLr_T-1qS0gZH75aKtMN3Yj0iPS4hcgUuTwjAzZr1Z9CAow";

static const char *jwk_sig =
  "dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk";

/*
 * Return the base64url encoded JWS signature of HEADER with PAYLOAD
 * (both already base64url encoded and with their lengths provided) using the
 * given symmetric key (KEY_BUF of KEY_LEN bytes).
 * Return NULL if an error occurs.
 */
Ds *
auth_oidc_jws_sign(const char *digest_name, unsigned char *key_buf,
				   unsigned int key_len,
				   const char *header, unsigned int header_len,
				   const char *payload, unsigned int payload_len)
{
  unsigned int hmac_len;
  unsigned char *hmac_buf;
  Ds *ds_sig;
  Hmac_handle *hmac;

  if ((hmac = crypto_hmac_open(digest_name, key_buf, key_len)) == NULL) {
	fprintf(stderr, "Cannot HMAC using %s\n", digest_name);
	return(NULL);
  }

  if (crypto_debug_flag) {
	dump_c(header, header_len);
	dump_c(".", 1);
	dump_c(payload, payload_len);
  }
  crypto_hmac_hash(hmac, (unsigned char *) header, header_len);
  crypto_hmac_hash(hmac, (unsigned char *) ".", 1);
  crypto_hmac_hash(hmac, (unsigned char *) payload, payload_len);

  hmac_buf = crypto_hmac_close(hmac, NULL, &hmac_len);
  if (crypto_debug_flag) {
	unsigned int mesg_len;

	mesg_len = header_len + 1 + payload_len;
	fprintf(stderr, "HMAC-%s (mesg_len=%u):\n", digest_name, mesg_len);
	dump_ub(hmac_buf, hmac_len);
  }

  ds_sig = ds_strba64_x(NULL, STRBA64_B64URL, hmac_buf, hmac_len);
  return(ds_sig);
}

#include "jspr/jspr.h"

/*
 * Using DIGEST_NAME, KEY_BUF, and KEY_LEN, compute and validate the
 * signature attached to JSON Web Signature structure JWS_MESG.
 * If successful, optionally return the base64url decoded header, payload,
 * and signature, and return 1.
 * If validation fails, only return -1.
 */
int
auth_oidc_jws_validate(const char *digest_name, unsigned char *key_buf,
					   unsigned int key_len, const char *jws_mesg,
					   Ds **ds_decoded_header, Ds **ds_decoded_payload,
					   Ds **ds_decoded_sig)
{
  const char *end_header, *end_payload, *end_sig;
  const char *start_header, *start_payload, *start_sig;
  unsigned int header_len, payload_len, sig_len;
  Ds *ds_calcsig, *ds_header, *ds_payload, *ds_sig;
  static int jws_validate_header_type = 0;

  start_header = jws_mesg;
  if ((end_header = strchr(start_header, (int) '.')) == NULL)
	return(-1);
  header_len = end_header - start_header;
  ds_header = ds_stra64b_x(NULL, STRA64B_B64URL, start_header, '.');
  if (ds_header == NULL) {
	fprintf(stderr, "Cannot decode header\n");
	return(-1);
  }
  ds_appendc(ds_header, (int) '\0');

  start_payload = end_header + 1;
  if ((end_payload = strchr(start_payload, (int) '.')) == NULL)
	return(-1);
  payload_len = end_payload - start_payload;
  ds_payload = ds_stra64b_x(NULL, STRA64B_B64URL, start_payload, '.');
  if (ds_payload == NULL) {
	fprintf(stderr, "Cannot decode payload\n");
	return(-1);
  }
  ds_appendc(ds_payload, (int) '\0');

  start_sig = end_payload + 1;
  sig_len = strlen(start_sig);
  ds_sig = ds_stra64b_x(NULL, STRA64B_B64URL, start_sig, 0);
  /* XXX also check for valid signature length. */
  if (ds_sig == NULL) {
	fprintf(stderr, "Cannot find valid signature\n");
	return(-1);
  }
  ds_appendc(ds_sig, (int) '\0');

  /*
   * Parse the header, primarily to extract the signature algorithm parameter,
   * but also as a superficial syntactical sanity check.
   */
  {
	char *alg, *typ;
	Jspr_parser *jp;
	Jspr_rc rc;
	Jspr_value *v;

	jp = jspr_init(ds_buf(ds_header), ds_len(ds_header) - 1);
	if ((rc = jspr_parse(jp)) != JSPR_OK) {
	  fprintf(stderr, "Cannot parse JWS\n");
	  return(-1);
	}

	/* The "typ" header parameter is optional. */
	if (jws_validate_header_type) {
	  if ((v = jspr_member_lookup(jp, NULL, "typ", JSPR_ANY_TYPE)) == NULL) {
		fprintf(stderr, "Cannot find JWT type parameter\n");
		return(-1);
	  }
	  typ = jspr_string_copy(jp, v);
	  if (typ == NULL) {
		fprintf(stderr, "Cannot find JWT type value\n");
		return(-1);
	  }
	  if (!streq(typ, "JWT")) {
		fprintf(stderr, "Invalid JWT type value: \"%s\"\n", typ);
		return(-1);
	  }
	}

 	if ((v = jspr_member_lookup(jp, NULL, "alg",  JSPR_ANY_TYPE)) == NULL)
	  return(-1);
	alg = jspr_string_copy(jp, v);
	if (alg == NULL) {
	  fprintf(stderr, "Cannot find JWS signature algorithm parameter\n");
	  return(-1);
	}
	if (!streq(alg, "HS256")) {
	  fprintf(stderr, "Invalid JWS alg: \"%s\"\n", alg);
	  return(-1);
	}

	jspr_end(jp);
  }

  /* Compute the expected signature and compare it with the one provided. */
  ds_calcsig = auth_oidc_jws_sign(digest_name, key_buf, key_len, 
								  start_header, header_len, start_payload,
								  payload_len);
  if (ds_calcsig == NULL) {
	fprintf(stderr, "Cannot compute JWS signature\n");
	return(-1);
  }

  if (!streq(ds_buf(ds_calcsig), start_sig)) {
	fprintf(stderr, "JWS signatures do not match\n");
	return(-1);
  }

  if (ds_decoded_header != NULL)
	*ds_decoded_header = ds_header;

  if (ds_decoded_payload != NULL)
	*ds_decoded_payload = ds_payload;

  if (ds_decoded_sig != NULL)
	*ds_decoded_sig = ds_sig;

  return(1);
}

int
auth_oidc_do_jws_test(void)
{
  int rc;
  unsigned int hmac_len, key_len, len;
  unsigned char *hmac_buf, *key_buf;
  Ds *ds_header, *ds_jws, *ds_key, *ds_payload, *ds_sep, *ds_sig;
  Hmac_handle *hmac;

  ds_key = ds_stra64b_x(NULL, STRA64B_B64URL, jwk_key_oct, 0);
  if (ds_key == NULL || (key_len = (unsigned int) ds_len(ds_key)) == 0) {
	fprintf(stderr, "test_jws: key decoding error\n");
	exit(1);
  }
  key_buf = (unsigned char *) ds_buf(ds_key);
  if (crypto_debug_flag) {
	fprintf(stderr, "JWS symmetric key length=%u:\n", key_len);
	dump_hb(key_buf, key_len);
  }

  ds_header = ds_strba64_x(NULL, STRBA64_B64URL,
						   (const unsigned char *) jws_protected_header,
						   strlen(jws_protected_header));
  if (ds_header == NULL) {
	fprintf(stderr, "test_jws: JWS Protected Header decode error\n");
	exit(1);
  }
  if (crypto_debug_flag)
	fprintf(stderr, "JWS Protected Header:\n%s\n", ds_buf(ds_header));

  ds_payload = ds_strba64_x(NULL, STRBA64_B64URL,
							(const unsigned char *) jws_payload,
							strlen(jws_payload));
  if (ds_payload == NULL) {
	fprintf(stderr, "test_jws: JWS Payload decode error\n");
	exit(1);
  }
  if (crypto_debug_flag)
	fprintf(stderr, "JWS Payload:\n%s\n", ds_buf(ds_payload));

  ds_sep = ds_setn(NULL, (unsigned char *) ".", 2);
  if (crypto_debug_flag) {
	Ds *ds;

	/* Concatenate three null-terminated strings, result not null-terminated. */
	ds = ds_dsappend(NULL, 1, 3, ds_header, ds_sep, ds_payload);
	ds_appendc(ds, (int) '\0');
	fprintf(stderr, "JWS Signing Input value:\n");
	fprintf(stderr, "%s\n", ds_buf(ds));
	if (crypto_debug_flag)
	  dump_ub(ds_buf(ds), ds_len(ds) - 1);
	fprintf(stderr, "\n");

	hmac_buf = crypto_hmac("SHA256", key_buf, key_len,
						   (unsigned char *) ds_buf(ds), ds_len(ds) - 1,
						   NULL, &hmac_len);
	fprintf(stderr, "HMAC-SHA256 (mesg_len=%lu):\n",
			(unsigned long) ds_len(ds) - 1);
	dump_ub(hmac_buf, hmac_len);
	fprintf(stderr, "\n");
  }

  ds_sig = auth_oidc_jws_sign("SHA256", key_buf, key_len,
							  ds_buf(ds_header), ds_len(ds_header) - 1,
							  ds_buf(ds_payload), ds_len(ds_payload) - 1);
  if (ds_sig != NULL)
	fprintf(stderr, "JWS Signature (base64url):\n%s\n", ds_buf(ds_sig));

  if (!streq(ds_buf(ds_sig), jwk_sig)) {
	fprintf(stderr, "Computed signature is incorrect\n");
	exit(1);
  }

  fprintf(stderr, "ok... Signature is correct\n");

  ds_jws = ds_dsappend(NULL, 1, 5, ds_header, ds_sep, ds_payload, ds_sep,
					   ds_sig);

  rc = auth_oidc_jws_validate("SHA256", key_buf, key_len, ds_buf(ds_jws),
						 NULL, NULL, NULL);
  if (rc != 1) {
	fprintf(stderr, "Computed signature did not validate!\n");
	exit(1);
  }
  fprintf(stderr, "ok... Signature validated\n");

  return(0);
}
