/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * This comprises a layer that maps OpenSSL to custom/alternate implementations
 * of cryptographic functionality using OpenSSL 1.1.1's engine mechanism.
 * Its purpose is only to give certain capabilities and functionality to
 * DACS, not to extend OpenSSL so that non-DACS applications might use them.
 *
 * For newer OpenSSL 3.X branches, refer to crypto_provider.c, which does
 * essentially the same thing.
 *
 * Because the ENGINE API is deprecated in OpenSSL 3.0, as part of the code
 * refactoring to (eventually) support that branch, code specific to the
 * OpenSSL ENGINE API was moved here from crypto.c for DACS 1.4.50 and
 * subsequent releases.
 *
 * Support for the OpenSSL 1.1.1 branch (and therefore the engine API) will
 * continue but at some point will be deprecated with adequate warning.
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2024\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: crypto_engine.c 3321 2024-06-19 23:39:55Z brachman $";
#endif

#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "local.h"
#include "crypto_sha512.h"
#include "dacs_crypto.h"

#if (OPENSSL_VERSION_NUMBER >= 0x30000000L)
#else

#include "crypto_engine.h"

static char *log_module_name = "crypto_engine";

#ifdef ENABLE_REQUIRE_OPENSSL_DIGESTS
static int require_openssl_digests = 1;
#else
static int require_openssl_digests = 0;
#endif

/*
 *
 */
typedef struct Engine_digest_tab Engine_digest_tab;
struct Engine_digest_tab {
  Digest_tab *dt;
  EVP_MD *evp;
  Engine_digest_tab *next;
};

/*
 * OpenSSL ENGINE configuration for a collection of digests.
 */
typedef struct Engine_config {
  char *engine_id;
  char *engine_name;
  ENGINE *engine;
  int nid_count;
  int *digest_nids;
  Engine_digest_tab *engine_digest_tab;
} Engine_config;

/*
 * Only this one engine is used, and just for a few
 * digests; these routines will need to be generalized if
 * additional functionality is required some day
 */
static Engine_config dacs_engine_config = {
  "DACS-Engine", "DACS OpenSSL engine", NULL, 0, NULL, NULL
};

/* For debugging only... */
static char *
to_hex(unsigned char *value, size_t nbytes)
{
  int i;
  char *p;
  unsigned char *v;
  static char buf[512];

  if (((nbytes * 2) + 1) > sizeof(buf))
	return(NULL);

  v = value;
  p = buf;
  for (i = 0; i < nbytes; i++)
	p += sprintf(p, "%.2x", *v++ & 0xff);

  return(buf);
}

static int
engine_copy_default(EVP_MD_CTX *to, const EVP_MD_CTX *from)
{

  return(1);
}

static int
engine_cleanup_default(EVP_MD_CTX *ctx)
{

  return(1);
}

/*
 * Return configuration for the digest configured for EC that is identified by
 * NID, or NULL.
 */
static Engine_digest_tab *
engine_lookup_digest_by_nid(Engine_config *ec, int nid)
{
  Engine_digest_tab *edt;

  if (nid == NID_undef)
	return(NULL);

  for (edt = ec->engine_digest_tab; edt != NULL; edt = edt->next) {
	if (edt->dt->ossl_nid == nid)
	  return(edt);
  }

  return(NULL);
}

/*
 * Return configuration for the digest configured for EC that is identified by
 * DIGEST_NAME, or NULL.
 */
static Engine_digest_tab *
engine_lookup_digest_by_name(Engine_config *ec, char *digest_name)
{
  Engine_digest_tab *edt;

  if (digest_name == NULL)
	return(NULL);

  for (edt = ec->engine_digest_tab; edt != NULL; edt = edt->next) {
	if (crypto_digest_name_match(edt->dt, digest_name))
	  return(edt);
  }

  return(NULL);
}

/*
 * Return 1 if OpenSSL engine E has been configured by engine_init(),
 * zero otherwise.
 */
static int
engine_is_configured(ENGINE *e)
{

  return((dacs_engine_config.engine == e) ? 1 : 0);
}

/*
 * Called by the OpenSSL engine layer to return information about custom
 * digest implementations.
 */
static int
engine_digests_callback(ENGINE *e, const EVP_MD **digest, const int **nids,
						int nid)
{
  int ok;
  Engine_digest_tab *edt;

  log_msg((LOG_DEBUG_LEVEL, "Entering engine_digests_callback()..."));

  if (!engine_is_configured(e))
	return(0);

  if (digest == NULL) {
	int n;

    /* Return a list of supported NIDs. */
    *nids = dacs_engine_config.digest_nids;
	n = dacs_engine_config.nid_count;
	log_msg((LOG_DEBUG_LEVEL, "Returning list of %d NIDs", n));

    return(n);
  }

  /* We are being queried for a specific digest, NID. */
  if ((edt = engine_lookup_digest_by_nid(&dacs_engine_config, nid)) != NULL) {
	log_msg((LOG_DEBUG_LEVEL, "Query nid: NID_%s (NID=%d)",
			 OBJ_nid2ln(nid), nid));
	*digest = edt->evp;
	ok = 1;
  }
  else {
	log_msg((LOG_DEBUG_LEVEL, "Query nid: NID%s (NID=%d)",
			 OBJ_nid2ln(nid), nid));
    *digest = NULL;
    ok = 0;
  }

  return(ok);
}

/*
 * A digest's init function will be called indirectly through the generic API;
 * e.g., EVP_DigestInit_ex(ctx, evp, engine)
 * Note that a direct call to MD5Init() seems to be hard coded to
 * OpenSSL's library function (i.e., it is not remapped by engine shenanigans)
 */
static int
engine_md_init_func(EVP_MD_CTX *ctx)
{
  int nid;
  const EVP_MD *md;

  md = EVP_MD_CTX_md(ctx);
  nid = EVP_MD_type(md);
  log_msg((LOG_TRACE_LEVEL, "md_init_func (NID%d, sz=%u, bs=%u)",
		   nid, EVP_MD_size(md), EVP_MD_CTX_block_size(ctx)));

  switch (nid) {
  case NID_md5:
	MD5_init((MD5_ctx *) EVP_MD_CTX_md_data(ctx));
	break;

  case NID_sha1:
	SHA1_init((SHA1_ctx *) EVP_MD_CTX_md_data(ctx));
	break;

  case NID_sha384:
	sha512_384_init((SHA512_ctx *) EVP_MD_CTX_md_data(ctx));
	break;

  case NID_sha512_224:
	sha512_224_init((SHA512_ctx *) EVP_MD_CTX_md_data(ctx));
	break;

  case NID_sha512_256:
	sha512_256_init((SHA512_ctx *) EVP_MD_CTX_md_data(ctx));
	break;

  case NID_sha512:
	sha512_init((SHA512_ctx *) EVP_MD_CTX_md_data(ctx));
	break;

  default:
	return(0);
  }

  return(1);
}

/*
 * Called via
 *  EVP_DigestUpdate(ctx, data, count)
 * but not MD5Update(ctx->md_data, data, count)
 */
static int
engine_md_update_func(EVP_MD_CTX *ctx, const void *data, size_t count)
{
  int nid;
  const EVP_MD *md;

  md = EVP_MD_CTX_md(ctx);
  nid = EVP_MD_type(md);
  log_msg((LOG_TRACE_LEVEL, "md_update_func NID%d", nid));

  switch (nid) {
  case NID_md5:
	MD5_update((MD5_ctx *) EVP_MD_CTX_md_data(ctx), data, count);
	break;

  case NID_sha1:
	SHA1_update((SHA1_ctx *) EVP_MD_CTX_md_data(ctx), data, count);
	break;

  case NID_sha384:
	sha512_update((SHA512_ctx *) EVP_MD_CTX_md_data(ctx), data, count);
	break;

  case NID_sha512_224:
	sha512_update((SHA512_ctx *) EVP_MD_CTX_md_data(ctx), data, count);
	  return(-1);
	break;

  case NID_sha512_256:
	sha512_update((SHA512_ctx *) EVP_MD_CTX_md_data(ctx), data, count);
	break;

  case NID_sha512:
	sha512_update((SHA512_ctx *) EVP_MD_CTX_md_data(ctx), data, count);
	break;

  default:
	return(0);
  }

  return(1);
}

/*
 * Called via
 *   EVP_DigestFinal(ctx, out, &digestSize);
 * but not MD5Final(md, ctx->md_data)
 */
static int
engine_md_final_func(EVP_MD_CTX *ctx, unsigned char *res)
{
  int nid;
  const char *name;
  const EVP_MD *md;

  md = EVP_MD_CTX_md(ctx);
  nid = EVP_MD_type(md);
  name = OBJ_nid2sn(EVP_MD_type(md));
  log_msg((LOG_TRACE_LEVEL, "md_final_func \"%s\" (NID%d)", name, nid));

  switch (nid) {
  case NID_md5:
	MD5_final((MD5_ctx *) EVP_MD_CTX_md_data(ctx), res);
	break;

  case NID_sha1:
	SHA1_final((SHA1_ctx *) EVP_MD_CTX_md_data(ctx), res);
	break;

  case NID_sha384:
	sha512_384_final((SHA512_ctx *) EVP_MD_CTX_md_data(ctx), res);
	break;

  case NID_sha512_224:
	sha512_224_final((SHA512_ctx *) EVP_MD_CTX_md_data(ctx), res);
	break;

  case NID_sha512_256:
	sha512_256_final((SHA512_ctx *) EVP_MD_CTX_md_data(ctx), res);
	break;

  case NID_sha512:
	sha512_final((SHA512_ctx *) EVP_MD_CTX_md_data(ctx), res);
	break;

  default:
	return(0);
  }

  return(1);
}

static int
engine_md_copy_func(EVP_MD_CTX *to_ctx, const EVP_MD_CTX *from_ctx)
{
  int nid;
  const char *name;
  const EVP_MD *md;

  md = EVP_MD_CTX_md(from_ctx);
  nid = EVP_MD_type(md);
  name = OBJ_nid2sn(EVP_MD_type(md));
  log_msg((LOG_TRACE_LEVEL, "md_copy_func \"%s\" (NID%d)", name, nid));

  switch (nid) {
  case NID_md5:
  case NID_sha1:
  case NID_sha384:
  case NID_sha512_224:
  case NID_sha512_256:
  case NID_sha512:
	break;

  default:
	return(0);
  }

  return(1);
}

static int
engine_md_cleanup_func(EVP_MD_CTX *ctx)
{
  int nid;
  const char *name;
  const EVP_MD *md;

  md = EVP_MD_CTX_md(ctx);
  nid = EVP_MD_type(md);
  name = OBJ_nid2sn(EVP_MD_type(md));
  log_msg((LOG_TRACE_LEVEL, "md_cleanup_func \"%s\" (NID%d)", name, nid));

  switch (nid) {
  case NID_md5:
  case NID_sha1:
  case NID_sha384:
  case NID_sha512_224:
  case NID_sha512_256:
  case NID_sha512:
	break;

  default:
	return(0);
  }

  return(1);
}

/*
 * Configure digest EDC so that it is callable through OpenSSL.
 * Return a new configured EVP_MD structure for digest EDC, or NULL if an
 * error occurs.
 */
static EVP_MD *
engine_digest_config(Digest_tab *dt)
{
  EVP_MD *evp;

  log_msg((LOG_DEBUG_LEVEL, "Configuring: NID_%s (NID==%d)",
		   OBJ_nid2ln(dt->ossl_nid), dt->ossl_nid));

  if ((evp = EVP_MD_meth_new(dt->ossl_nid, NID_undef)) == NULL) {
	log_msg((LOG_ERROR_LEVEL, "EVP_MD_meth_new() failed"));

  err:
	return(NULL);
  }

  if (EVP_MD_meth_set_result_size(evp, dt->digest_size) == 0) {
	log_msg((LOG_ERROR_LEVEL, "EVP_MD_meth_set_result_size() failed"));
	goto err;
  }

  /*
   * If the app_datasize is non-zero,
   * a subsequent call to EVP_DigestInit() [or EVP_DigestInit_ex()] will
   * allocate private memory for the digest's context within the generic
   * EVP_MD_CTX, unless the EVP_MD_CTX_FLAG_NO_INIT is passed; e.g.,
   *   ctx->md_data = (MD5_ctx *) malloc(sizeof(MD5_ctx));
   * When EVP_DigestFinal() [or EVP_DigestFinal_ex()] is called, this allocation
   * will be freed.
   */
  if (EVP_MD_meth_set_app_datasize(evp, dt->app_datasize) == 0) {
	log_msg((LOG_ERROR_LEVEL, "EVP_MD_meth_set_app_datasize() failed"));
	goto err;
  }

  /*
   * Instead of having five callback functions per configured digest
   * (i.e., init, update, final, copy, cleanup), which quickly
   * becomes a large mess, we will use five generic callback functions, each of
   * which determines at runtime the digest function that needs to be called.
   * This is slightly less runtime efficient, but smaller and easier to work with.
   */
  if (EVP_MD_meth_set_init(evp, engine_md_init_func) == 0) {
	log_msg((LOG_ERROR_LEVEL, "EVP_MD_meth_set_init() failed"));
	goto err;
  }
  if (crypto_debug_flag)
	fprintf(stderr, "set_init to engine_md_init_func: NID%d\n", dt->ossl_nid);

  if (EVP_MD_meth_set_update(evp, engine_md_update_func) == 0) {
	log_msg((LOG_ERROR_LEVEL, "EVP_MD_meth_set_update() failed"));
	goto err;
  }
  if (crypto_debug_flag)
	fprintf(stderr, "set_update to engine_md_update_func: NID%d\n", dt->ossl_nid);

  if (EVP_MD_meth_set_final(evp, engine_md_final_func) == 0) {
	log_msg((LOG_ERROR_LEVEL, "EVP_MD_meth_set_final() failed"));
	goto err;
  }
  if (crypto_debug_flag)
	fprintf(stderr, "set_final to engine_md_final_func: NID%d\n", dt->ossl_nid);

  if (EVP_MD_meth_set_copy(evp, engine_md_copy_func) == 0) {
	log_msg((LOG_ERROR_LEVEL, "EVP_MD_meth_set_copy() failed"));
	goto err;
  }

  if (EVP_MD_meth_set_cleanup(evp, engine_md_cleanup_func) == 0) {
	log_msg((LOG_ERROR_LEVEL, "EVP_MD_meth_set_cleanup() failed"));
	goto err;
  }

  return(evp);
}

static int
engine_add_digest(Engine_config *ec, Digest_tab *dt)
{
  Engine_digest_tab *d, *edt;

  if (dt->ossl_nid == NID_undef) {
	log_msg((LOG_ERROR_LEVEL, "Attempt to add invalid digest \"%s\" to engine",
			 dt->name));
	return(-1);
  }

  d = (Engine_digest_tab *) MALLOC(sizeof(Engine_digest_tab));
  d->dt = dt;
  d->evp = engine_digest_config(dt);
  d->next = ec->engine_digest_tab;
  ec->engine_digest_tab = d;

  return(0);
}

/*
 * Create and initialize an OpenSSL engine as described by EC.
 *
 * Whether or not OpenSSL's engine API is used for a particular algorithm
 * depends on configuration done elsewhere, e.g., such as the DIGEST_USE_ENGINE
 * flag in digest_tab.  This flexibility is unlikely to be important for the
 * average deployment of DACS, but could be critical if OpenSSL does not
 * provide an algorithm DACS requires, or if we need to work around a bug
 * or performance issue in a third-party implementation of a cryptographic
 * function.
 *
 * The impetus for introducing this engine "switch" layer was the lack of
 * MD5 and SHA1 support in the FIPS-mode OpenSSL crypto library.
 * It appears that FIPS-mode OpenSSL is still "aware" of MD5 and SHA1, in that
 * NID_md5 and NID_sha1 are defined.
 *
 * Return the new ENGINE, or NULL if an error occurs.
 */
static ENGINE *
engine_init(Engine_config *ec, Digest_tab *digest_tab)
{
  int i, rc;
  Digest_tab *dt;
  Engine_digest_tab *edt;
  ENGINE *engine;

  log_msg((LOG_DEBUG_LEVEL, "Entering engine_init()..."));

  rc = -1;
  if ((engine = ENGINE_new()) == NULL) {
    log_msg((LOG_TRACE_LEVEL, "Engine_new() returned NULL"));
	goto end;
  }

  /*
   * XXX duplication using EVP_MD_meth_dup() probably won't work if the
   * method is unavailable, so initialize fully.
   */

  /*
   * Count the number of NIDs (digests) supported by the engine and
   * create a vector of their NIDs so they can be identified through
   * digests_callback(). The vector is terminated by a zero element.
   */
  ec->nid_count = 0;
  for (dt = digest_tab; dt->name != NULL; dt++) {
	if (digest_flag_is_enabled(dt->flags, DIGEST_USE_ENGINE)) {
	  if (engine_add_digest(&dacs_engine_config, dt) == -1)
		goto end;
	  ec->nid_count++;
	}
  }

  ec->digest_nids = (int *) MALLOC(sizeof(int) * (ec->nid_count + 1));
  edt = ec->engine_digest_tab;
  for (i = 0; edt != NULL; edt = edt->next, i++) {
	ec->digest_nids[i] = edt->dt->ossl_nid;
	edt = edt->next;
  }
  ec->digest_nids[i] = 0;

  if (!ENGINE_set_id(engine, ec->engine_id)) {
    log_msg((LOG_ERROR_LEVEL, "ENGINE_set_id() failed"));
    goto end;
  }

  if (!ENGINE_set_name(engine, ec->engine_name)) {
    log_msg((LOG_ERROR_LEVEL, "ENGINE_set_name() failed"));
    goto end;
  }

  if (!ENGINE_set_digests(engine, engine_digests_callback)) {
    log_msg((LOG_ERROR_LEVEL, "ENGINE_set_digests() failed"));
    goto end;
  }

  if (!ENGINE_set_default(engine, ENGINE_METHOD_DIGESTS)) {
    log_msg((LOG_ERROR_LEVEL, "ENGINE_set_default() failed"));
    goto end;
  }

  if (!ENGINE_add(engine)) {
    log_msg((LOG_ERROR_LEVEL, "ENGINE_add() failed"));
    goto end;
  }

  if (!ENGINE_init(engine))
	goto end;

  ec->engine = engine;
  rc = 0;

 end:

  if (rc == -1) {
	crypto_log_error();
	log_msg((LOG_ERROR_LEVEL, "engine_init() Error"));
	if (engine != NULL)
	  ENGINE_free(engine);
  }
  else
	log_msg((LOG_DEBUG_LEVEL, "engine_init() OK"));

  return(engine);
}

static MAYBE_UNUSED void
engine_free(ENGINE *engine)
{

  log_msg((LOG_DEBUG_LEVEL, "engine_free()"));

  if (engine != NULL) {
	ENGINE_cleanup();
	ENGINE_finish(engine);
	ENGINE_free(engine);
  }
}

int
crypto_engine_init(Digest_tab *digest_tab)
{
  ENGINE *e;

#if (OPENSSL_VERSION_NUMBER >= 0x30000000L)
  return(-1);
#else
#ifdef CRYPT_ENGINE_NAME
  {
	if ((e = ENGINE_by_id(CRYPT_ENGINE_NAME)) == NULL) {
	  log_msg((LOG_ERROR_LEVEL, "ENGINE_by_id lookup error for %s",
			   CRYPT_ENGINE_NAME));
	  return(-1);
	}

	if (ENGINE_set_default(e, ENGINE_METHOD_ALL) == 0) {
	  log_msg((LOG_ERROR_LEVEL, "ENGINE_set_default error for %s",
			   CRYPT_ENGINE_NAME));
	  return(-1);
	}
  }
#endif

  /*
   * Force either OpenSSL's or DACS's use of certain digest functions.
   * This is currently a build-time choice.
   */
  if (crypto_require_openssl_digests(require_openssl_digests) == -1) {
	log_msg((LOG_ERROR_LEVEL, "Digest initialization snafu... fix me"));
	return(-1);
  }
  log_msg((LOG_DEBUG_LEVEL, "require_openssl_digests=%d",
		   require_openssl_digests));
#ifdef PROG
  log_msg((LOG_DEBUG_LEVEL,
		   "%s OpenSSL+DACS engine-based code [require_openssl_digests=%d]",
		   (require_openssl_digests == 1) ? "Not using" : "Using",
		   require_openssl_digests));
#endif

  /* Make all built-in OpenSSL ENGINE implementations available. */
  ENGINE_load_builtin_engines();
  log_msg((LOG_DEBUG_LEVEL, "Loaded built-in engines."));

  /* See: OPENSSL_init_crypto(OPENSSL_INIT_ENGINE_DYNAMIC, NULL) */
  ENGINE_load_dynamic();

  if ((e = engine_init(&dacs_engine_config, digest_tab)) == NULL) {
	log_msg((LOG_ERROR_LEVEL, "Could not initialize engine: \"%s\"",
			 dacs_engine_config.engine_id));
	crypto_log_error();
	return(-1);
  }

  log_msg((LOG_DEBUG_LEVEL, "Successfully initialized engine: \"%s\" (\"%s\")",
		   dacs_engine_config.engine_id, dacs_engine_config.engine_name));

  return(0);
#endif
}

/*
 *
 */
EVP_MD *
crypto_engine_make_ctx(Digest_tab *dt)
{
  Engine_digest_tab *edt;

#if (OPENSSL_VERSION_NUMBER >= 0x30000000L)
  return(NULL);
#else

  edt = engine_lookup_digest_by_nid(&dacs_engine_config, dt->ossl_nid);

  if (edt == NULL) {
	log_msg((LOG_ERROR_LEVEL,
			 "Internal error: Engine digest \"%s\" is not implemented",
			 dt->name));
	return(NULL);
  }

  return(edt->evp);
#endif
}

static int
engine_test_sha512(ENGINE *engine, char *str, unsigned char *out)
{
  int er;
  unsigned int digestSize = -1;
  Engine_digest_tab *edt;
  EVP_MD_CTX *ctx;

  fprintf(stderr, "Testing SHA512(\"%s\")... ", str);
  edt = engine_lookup_digest_by_nid(&dacs_engine_config, NID_sha512);
  if (edt == NULL) {
  fail:
	fprintf(stderr, "Failed\n");
	return(-1);
  }

  ctx = EVP_MD_CTX_new();

  if ((er = EVP_DigestInit_ex(ctx, edt->evp, engine)) == 0)
	goto fail;

  if ((er = EVP_DigestUpdate(ctx, str, strlen(str))) == 0)
	goto fail;

  if ((er = EVP_DigestFinal(ctx, out, &digestSize)) == 0)
	goto fail;

  /* Renamed to EVP_MD_CTX_free() in OpenSSL 1.1.0 */
  EVP_MD_CTX_free(ctx);

  fprintf(stderr, "OK\n");
  return(0);
}

static int
engine_test_md5(ENGINE *engine, char *str, unsigned char *out)
{
  int er;
  unsigned int digestSize = -1;
  Engine_digest_tab *edt;
  EVP_MD_CTX *ctx;

  fprintf(stderr, "Testing MD5(\"%s\")... ", str);
  edt = engine_lookup_digest_by_nid(&dacs_engine_config, NID_md5);
  if (edt == NULL) {
  fail:
	fprintf(stderr, "Failed\n");
	return(-1);
  }

  ctx = EVP_MD_CTX_new();

  if ((er = EVP_DigestInit_ex(ctx, edt->evp, engine)) == 0)
	goto fail;

  if ((er = EVP_DigestUpdate(ctx, str, strlen(str))) == 0)
	goto fail;

  if ((er = EVP_DigestFinal(ctx, out, &digestSize)) == 0)
	goto fail;

  /* Renamed to EVP_MD_CTX_free() in OpenSSL 1.1.0 */
  EVP_MD_CTX_free(ctx);

  fprintf(stderr, "OK\n");
  return(0);
}

static int
engine_test_sha1(ENGINE *engine, char *str, unsigned char *out)
{
  int er;
  unsigned int digestSize = -1;
  Engine_digest_tab *edt;
  EVP_MD_CTX *ctx;

  fprintf(stderr, "Testing SHA1(\"%s\")... ", str);
  edt = engine_lookup_digest_by_nid(&dacs_engine_config, NID_sha1);
  if (edt == NULL) {
  fail:
	fprintf(stderr, "Fail\n");
	return(-1);
  }

  /* Renamed to EVP_MD_CTX_new() in OpenSSL 1.1.0 */
  ctx = EVP_MD_CTX_new();

  if ((er = EVP_DigestInit_ex(ctx, edt->evp, engine)) == 0)
	goto fail;

  if ((er = EVP_DigestUpdate(ctx, str, strlen(str))) == 0)
	goto fail;

  if ((er = EVP_DigestFinal(ctx, out, &digestSize)) == 0)
	goto fail;

  /* Renamed to EVP_MD_CTX_free() in OpenSSL 1.1.0 */
  EVP_MD_CTX_free(ctx);

  fprintf(stderr, "OK\n");
  return(0);
}

static int
engine_test_sign(ENGINE *engine)
{
  Engine_digest_tab *edt;
  EVP_MD_CTX *ctx;

  fprintf(stderr, "Testing engine signing... ");
  edt = engine_lookup_digest_by_nid(&dacs_engine_config, NID_sha1);
  if (edt == NULL) {
  fail:
	fprintf(stderr, "Fail\n");
	return(-1);
  }

  /* Renamed to EVP_MD_CTX_new() in OpenSSL 1.1.0 */
  ctx = EVP_MD_CTX_new();

  if (!EVP_DigestInit_ex(ctx, edt->evp, engine))
	goto fail;

  if (!EVP_SignInit_ex(ctx, edt->evp, NULL))
	goto fail;

  if (!EVP_SignUpdate(ctx, "abc", 3))
	goto fail;

  /* Renamed to EVP_MD_CTX_free() in OpenSSL 1.1.0 */
  EVP_MD_CTX_free(ctx);

  return(0);
}

int
crypto_engine_tests(void)
{
  int err;
  unsigned int flags;

#if (OPENSSL_VERSION_NUMBER >= 0x30000000L)
  return(-1);
#else
  err = crypto_get_digest_flags("sha512", &flags);
  if (err == -1) {
	fprintf(stderr, "sha512 lookup failed!\n");
	return(-1);
  }

  if (digest_flag_is_enabled(flags, DIGEST_USE_ENGINE)) {
	unsigned char out[SHA512_DIGEST_BYTE_LENGTH];

	err = engine_test_sha512(dacs_engine_config.engine, "abc", out);
	if (err == -1)
	  return(-1);

	fprintf(stderr, "SHA512: %s\n", to_hex(out, SHA512_DIGEST_BYTE_LENGTH));
  }
  else
	fprintf(stderr, "SHA512 engine test is disabled\n");

  err = crypto_get_digest_flags("md5", &flags);
  if (err == -1) {
	fprintf(stderr, "md5 lookup failed!\n");
	return(-1);
  }

  if (digest_flag_is_enabled(flags, DIGEST_USE_ENGINE)) {
	unsigned char out[MD5_DIGEST_LEN];

	err = engine_test_md5(dacs_engine_config.engine, "abc", out);
	if (err == -1)
	  return(-1);

	fprintf(stderr, "MD5: %s\n", to_hex(out, MD5_DIGEST_LEN));
  }
  else
	fprintf(stderr, "MD5 engine test is disabled\n");

  err = crypto_get_digest_flags("sha1", &flags);
  if (err == -1) {
	fprintf(stderr, "sha1 lookup failed!\n");
	return(-1);
  }

  if (digest_flag_is_enabled(flags, DIGEST_USE_ENGINE)) {
	unsigned char out[SHA1_DIGEST_LEN];

	err = engine_test_sha1(dacs_engine_config.engine, "abc", out);
	if (err == -1)
	  return(-1);

	fprintf(stderr, "SHA1: %s\n", to_hex(out, SHA1_DIGEST_LEN));

	err = engine_test_sign(dacs_engine_config.engine);
	if (err == -1)
	  fprintf(stderr, "engine_test_sign() failed\n");
  }
  else
	fprintf(stderr, "SHA1 engine tests are disabled\n");

  return(err);
#endif
}

void
crypto_engine_end(void)
{

    engine_free(dacs_engine_config.engine);
}
#endif
