/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * Support for the OpenSSL 3.0 branch provider API, implemented as an
 * application built-in (i.e., it is not dynamically loaded).
 *
 * Refer to: provider(7)
 *  https://www.openssl.org/docs/man3.1/man7/provider.html
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2024\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: crypto_provider.c 3324 2024-06-20 22:14:04Z brachman $";
#endif

#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "local.h"
#include "dacs_crypto.h"

#if (OPENSSL_VERSION_NUMBER >= 0x30000000L)

#include <openssl/core.h>
#include <openssl/core_dispatch.h>
#include <openssl/core_names.h>
#include <openssl/provider.h>

#include "crypto_provider.h"

static int require_provider = 1;

static char *log_module_name = "crypto_provider";

/*
 * Function signatures for function pointers in OSSL_DISPATCH arrays
 * (used to catch incorrect function types in the dispatch table).
 */

/* Message digest context management and computation. */
static OSSL_FUNC_digest_newctx_fn md5_dacs_newctx;
static OSSL_FUNC_digest_freectx_fn md5_dacs_freectx;
static OSSL_FUNC_digest_init_fn md5_dacs_init;
static OSSL_FUNC_digest_update_fn md5_dacs_update;
static OSSL_FUNC_digest_final_fn md5_dacs_final;
static OSSL_FUNC_digest_digest_fn md5_dacs_digest;
static OSSL_FUNC_digest_get_params_fn md5_dacs_get_params;

/* Message digest context management and computation. */
static OSSL_FUNC_digest_newctx_fn sha1_dacs_newctx;
static OSSL_FUNC_digest_freectx_fn sha1_dacs_freectx;
static OSSL_FUNC_digest_init_fn sha1_dacs_init;
static OSSL_FUNC_digest_update_fn sha1_dacs_update;
static OSSL_FUNC_digest_final_fn sha1_dacs_final;
static OSSL_FUNC_digest_digest_fn sha1_dacs_digest;
static OSSL_FUNC_digest_get_params_fn sha1_dacs_get_params;

/* Provider initialization and termination. */
OSSL_FUNC_provider_query_operation_fn md5_dacs_query;
OSSL_FUNC_provider_teardown_fn md5_dacs_teardown;

/* For debugging only... */
static char *
to_hex(unsigned char *value, size_t nbytes)
{
  int i;
  char *p;
  unsigned char *v;
  static char buf[512];

  if (((nbytes * 2) + 1) > sizeof(buf))
	return(NULL);

  v = value;
  p = buf;
  for (i = 0; i < nbytes; i++)
	p += sprintf(p, "%.2x", *v++ & 0xff);

  return(buf);
}

/*****************************************************************************/
static void *
md5_dacs_newctx(void *provctx)
{
  MD5_ctx *ctx;

  ctx = ALLOC(MD5_ctx);

  return(ctx);
}

static void *
md5_dacs_dupctx(void *ctx)
{
  MD5_ctx *in, *out;

  in = (MD5_ctx *) ctx;
  out = (MD5_ctx *) ALLOC(MD5_ctx);

  *out = *in;
  return(out);
}

static void
md5_dacs_freectx(void *mctx)
{
  MD5_ctx *ctx;

  ctx = (MD5_ctx *) mctx;
  memzap(ctx, sizeof(MD5_ctx));
  free(ctx);
}

static int
md5_dacs_init(void *dctx, const OSSL_PARAM params[])
{
  MD5_ctx *ctx;

  if (!require_provider)
	return(-1);

  ctx = (MD5_ctx *) dctx;
  MD5_init(ctx);

  if (crypto_debug_flag)
	fprintf(stderr, "md5_dacs_init\n");

  return(1);
}

static int
md5_dacs_update(void *dctx, const unsigned char *in, size_t inl)
{
  MD5_ctx *ctx;

  if (!require_provider)
	return(-1);

  ctx = (MD5_ctx *) dctx;
  MD5_update(ctx, in, inl);

  if (crypto_debug_flag)
	fprintf(stderr, "md5_dacs_update\n");

  return(1);
}

static int
md5_dacs_final(void *dctx, unsigned char *out, size_t *outl, size_t outsize)
{
  MD5_ctx *ctx;

  ctx = (MD5_ctx *) dctx;
  MD5_final(ctx, out);
  if (outl != NULL)
	*outl = outsize;

  if (crypto_debug_flag)
	fprintf(stderr, "md5_dacs_final\n");
  return(1);
}

static int
md5_dacs_digest(void *provctx, const unsigned char *in, size_t inl,
				unsigned char *out, size_t *outl, size_t outsize)
{
  MD5_ctx ctx;

  MD5_init(&ctx);
  MD5_update(&ctx, (uint8_t *) in, inl);
  MD5_final(&ctx, out);
  if (outl != NULL)
	*outl = outsize;

  return(1);
}

/*
 * These two digest parameters must be provided.
 */
static int
md5_dacs_get_params(OSSL_PARAM params[])
{
  OSSL_PARAM *p = NULL;

  p = OSSL_PARAM_locate(params, OSSL_DIGEST_PARAM_BLOCK_SIZE);
  if (p != NULL && !OSSL_PARAM_set_size_t(p, MD5_DIGEST_LEN))
	return(0);

  p = OSSL_PARAM_locate(params, OSSL_DIGEST_PARAM_SIZE);
  if (p != NULL && !OSSL_PARAM_set_size_t(p, MD5_DIGEST_LEN))
	return(0);

  return(1);
}

/*****************************************************************************/
static void *
sha1_dacs_newctx(void *provctx)
{
  SHA1_ctx *ctx;

  ctx = ALLOC(SHA1_ctx);

  return(ctx);
}

static void *
sha1_dacs_dupctx(void *ctx)
{
  SHA1_ctx *in, *out;

  in = (SHA1_ctx *) ctx;
  out = (SHA1_ctx *) ALLOC(SHA1_ctx);

  *out = *in;
  return(out);
}

static void
sha1_dacs_freectx(void *mctx)
{
  SHA1_ctx *ctx;

  ctx = (SHA1_ctx *) mctx;
  memzap(ctx, sizeof(SHA1_ctx));
  free(ctx);
}

static int
sha1_dacs_init(void *dctx, const OSSL_PARAM params[])
{
  SHA1_ctx *ctx;

  if (!require_provider)
	return(-1);

  ctx = (SHA1_ctx *) dctx;
  SHA1_init(ctx);

  if (crypto_debug_flag)
	fprintf(stderr, "sha1_dacs_init\n");

  return(1);
}

static int
sha1_dacs_update(void *dctx, const unsigned char *in, size_t inl)
{
  SHA1_ctx *ctx;

  if (!require_provider)
	return(-1);

  ctx = (SHA1_ctx *) dctx;
  SHA1_update(ctx, in, inl);

  if (crypto_debug_flag)
	fprintf(stderr, "sha1_dacs_update\n");

  return(1);
}

static int
sha1_dacs_final(void *dctx, unsigned char *out, size_t *outl, size_t outsize)
{
  SHA1_ctx *ctx;

  ctx = (SHA1_ctx *) dctx;
  SHA1_final(ctx, out);
  if (outl != NULL)
	*outl = outsize;

  if (crypto_debug_flag)
	fprintf(stderr, "sha1_dacs_final\n");
  return(1);
}

static int
sha1_dacs_digest(void *provctx, const unsigned char *in, size_t inl,
				 unsigned char *out, size_t *outl, size_t outsize)
{
  SHA1_ctx ctx;

  SHA1_init(&ctx);
  SHA1_update(&ctx, (uint8_t *) in, inl);
  SHA1_final(&ctx, out);
  if (outl != NULL)
	*outl = outsize;

  return(1);
}

/*
 * These two digest parameters must be provided.
 */
static int
sha1_dacs_get_params(OSSL_PARAM params[])
{
  OSSL_PARAM *p = NULL;

  p = OSSL_PARAM_locate(params, OSSL_DIGEST_PARAM_BLOCK_SIZE);
  if (p != NULL && !OSSL_PARAM_set_size_t(p, SHA1_DIGEST_LEN))
	return(0);

  p = OSSL_PARAM_locate(params, OSSL_DIGEST_PARAM_SIZE);
  if (p != NULL && !OSSL_PARAM_set_size_t(p, SHA1_DIGEST_LEN))
	return(0);

  return(1);
}

/*****************************************************************************/
static OSSL_LIB_CTX *dacs_md5_libctx = NULL;
static OSSL_PROVIDER *dacs_md5_prov = NULL;

/* Functions we provide to the core. */
static const OSSL_DISPATCH prov_fns[] = {
  { OSSL_FUNC_PROVIDER_QUERY_OPERATION, (void (*)(void)) md5_dacs_query },
  { OSSL_FUNC_PROVIDER_TEARDOWN,        (void (*)(void)) md5_dacs_teardown },
  { 0, NULL }
};

/*
 * The provider initialization function, called by OpenSSL.
 */
static int
md5_dacs_provider_init(const OSSL_CORE_HANDLE *handle,
                       const OSSL_DISPATCH *in, const OSSL_DISPATCH **out,
					   void **provctx)
{

  if ((dacs_md5_libctx = OSSL_LIB_CTX_new()) == NULL)
	return(0);

  *provctx = dacs_md5_libctx;
  *out = prov_fns;

  return(1);
}

void
md5_dacs_teardown(void *provctx)
{

  if (provctx != NULL)
	free(provctx);
}

/*
 * Called by crypto_init() to initialize the DACS OpenSSL Provider to
 * supply digest algorithms.
 */
int
crypto_provider_init(void)
{

  if (!OSSL_PROVIDER_add_builtin(NULL, "dacs", md5_dacs_provider_init)) {
	dacs_fatal(ds_xprintf("Could not add builtin provider for md5"));
	exit(1);
  }

  if ((dacs_md5_prov = OSSL_PROVIDER_load(NULL, "dacs")) == NULL) {
	dacs_fatal(ds_xprintf("Could not load provider for DACS"));
	exit(1);
  }

  /* Make sure the OpenSSL default Provider is available. */
  if (OSSL_PROVIDER_load(NULL, "default") == NULL) {
	dacs_fatal(ds_xprintf("Could not load default provider"));
	exit(1);
  }

  log_msg((LOG_DEBUG_LEVEL, "Successfully initialized providers"));

  return(1);
}

void
crypto_provider_end(void)
{

  if (dacs_md5_prov != NULL)
	OSSL_PROVIDER_unload(dacs_md5_prov);
}

/*
 * See:
 *  https://www.openssl.org/docs/man3.1/man3/OSSL_DISPATCH.html
 *  openssl/core_dispatch.h
 */
static const OSSL_DISPATCH md5_dacs_functions[] = {
  { OSSL_FUNC_DIGEST_NEWCTX,     (void (*)(void)) md5_dacs_newctx },
  { OSSL_FUNC_DIGEST_DUPCTX,     (void (*)(void)) md5_dacs_dupctx },
  { OSSL_FUNC_DIGEST_FREECTX,    (void (*)(void)) md5_dacs_freectx },
  { OSSL_FUNC_DIGEST_INIT,       (void (*)(void)) md5_dacs_init },
  { OSSL_FUNC_DIGEST_UPDATE,     (void (*)(void)) md5_dacs_update },
  { OSSL_FUNC_DIGEST_FINAL,      (void (*)(void)) md5_dacs_final },
  { OSSL_FUNC_DIGEST_GET_PARAMS, (void (*)(void)) md5_dacs_get_params },
  { OSSL_FUNC_DIGEST_DIGEST,     (void (*)(void)) md5_dacs_digest },
  { 0, NULL }
};

static const OSSL_DISPATCH sha1_dacs_functions[] = {
  { OSSL_FUNC_DIGEST_NEWCTX,     (void (*)(void)) sha1_dacs_newctx },
  { OSSL_FUNC_DIGEST_DUPCTX,     (void (*)(void)) sha1_dacs_dupctx },
  { OSSL_FUNC_DIGEST_FREECTX,    (void (*)(void)) sha1_dacs_freectx },
  { OSSL_FUNC_DIGEST_INIT,       (void (*)(void)) sha1_dacs_init },
  { OSSL_FUNC_DIGEST_UPDATE,     (void (*)(void)) sha1_dacs_update },
  { OSSL_FUNC_DIGEST_FINAL,      (void (*)(void)) sha1_dacs_final },
  { OSSL_FUNC_DIGEST_GET_PARAMS, (void (*)(void)) sha1_dacs_get_params },
  { OSSL_FUNC_DIGEST_DIGEST,     (void (*)(void)) sha1_dacs_digest },
  { 0, NULL }
};

static const OSSL_PARAM md5_dacs_gettable_params[] = {
  OSSL_PARAM_size_t(OSSL_DIGEST_PARAM_BLOCK_SIZE, NULL),
  OSSL_PARAM_size_t(OSSL_DIGEST_PARAM_SIZE, NULL),
  OSSL_PARAM_END
};

/*
 * See:
 *  https://www.openssl.org/docs/man3.1/man3/OSSL_ALGORITHM.html
 */
static const OSSL_ALGORITHM md5_dacs_digest_alg[] = {
  { "md5",  "dacs.md5=yes",  md5_dacs_functions,  "DACS MD5 digest" },
  { "sha1", "dacs.sha1=yes", sha1_dacs_functions, "DACS SHA1 digest" },
  { NULL, NULL, NULL, NULL }
};

/*
 * Must be called to report the operations supplied by this provider.
 */
const OSSL_ALGORITHM *
md5_dacs_query(void *provctx, int operation_id, int *no_cache)
{

  *no_cache = 0;
  switch (operation_id) {
  case OSSL_OP_DIGEST:
	return(md5_dacs_digest_alg);
	break;

  default:
	break;
  }

  return(NULL);
}

Digest_ctx *
crypto_provider_make_ctx(Digest_tab *dt)
{
  Digest_ctx *ctx;
  EVP_MD *md;
  EVP_MD_CTX *evp_ctx;

  if ((md = EVP_MD_fetch(NULL, dt->name, dt->properties)) == NULL) {
	crypto_log_error();
	dacs_fatal(ds_xprintf("Internal error: Provider digest \"%s\" fetch failed",
						   dt->name));
	/*NOTREACHED*/
  }

  evp_ctx = EVP_MD_CTX_new();

  if (!EVP_DigestInit_ex2(evp_ctx, md, NULL)) {
	crypto_log_error();
	dacs_fatal(ds_xprintf("Internal error: Unable to initialize digest \"%s\"",
						  dt->name));
	/*NOTREACHED*/
  }

  ctx = ALLOC(Digest_ctx);
  ctx->dt = dt;

  ctx->u.evp_ctx = evp_ctx;
  dt->evp = md;

  return(ctx);
}

static int
provider_test_md5(void *ptr, char *str, unsigned char *out)
{
  int er;
  unsigned int digestSize = -1;
  EVP_MD *md;
  EVP_MD_CTX *ctx;

  fprintf(stderr, "Testing MD5(\"%s\")... ", str);

  if ((md = EVP_MD_fetch(NULL, "md5", "dacs.md5=yes")) == NULL) {
	fprintf(stderr, "EVP_MD_fetch failed\n");
	return(-1);
  }

  ctx = EVP_MD_CTX_new();

  /*
   * EVP_DigestInit(), EVP_DigestFinal(), and EVP_MD_CTX_copy() are obsolete.
   */
  if ((er = EVP_DigestInit_ex2(ctx, md, NULL)) == 0)
	goto fail;

  if ((er = EVP_DigestUpdate(ctx, str, strlen(str))) == 0)
	goto fail;

  if ((er = EVP_DigestFinal_ex(ctx, out, &digestSize)) == 0)
	goto fail;

  fprintf(stderr, "%s\n", to_hex(out, digestSize));
  EVP_MD_CTX_free(ctx);
  EVP_MD_free(md);

  return(0);

fail:
  return(-1);
}

static int
provider_test_sha1(void *ptr, char *str, unsigned char *out)
{
  int er;
  unsigned int digestSize = -1;
  EVP_MD *md;
  EVP_MD_CTX *ctx;

  fprintf(stderr, "Testing SHA1(\"%s\")... ", str);

  if ((md = EVP_MD_fetch(NULL, "sha1", "dacs.sha1=yes")) == NULL) {
	fprintf(stderr, "EVP_MD_fetch failed\n");
	return(-1);
  }

  ctx = EVP_MD_CTX_new();

  if ((er = EVP_DigestInit_ex2(ctx, md, NULL)) == 0)
	goto fail;

  if ((er = EVP_DigestUpdate(ctx, str, strlen(str))) == 0)
	goto fail;

  if ((er = EVP_DigestFinal(ctx, out, &digestSize)) == 0)
	goto fail;

  fprintf(stderr, "%s\n", to_hex(out, digestSize));
  EVP_MD_CTX_free(ctx);
  EVP_MD_free(md);

  return(0);

fail:
  return(-1);
}

static int
provider_test_sign(void *ptr)
{
  EVP_MD *md;
  EVP_MD_CTX *ctx;

  fprintf(stderr, "Testing provider signing... ");

  if ((md = EVP_MD_fetch(NULL, "sha1", "dacs.sha1=yes")) == NULL) {
	fprintf(stderr, "EVP_MD_fetch failed\n");
	return(-1);
  }

  ctx = EVP_MD_CTX_new();

  if (!EVP_DigestInit_ex(ctx, md, NULL))
	goto fail;

  if (!EVP_SignInit_ex(ctx, md, NULL))
	goto fail;

  if (!EVP_SignUpdate(ctx, "abc", 3))
	goto fail;

  EVP_MD_CTX_free(ctx);

  return(0);

 fail:
  return(-1);
}

int
crypto_provider_tests(void)
{
  int err;
  unsigned int flags;
  unsigned char out[MD5_DIGEST_LEN];

  err = crypto_get_digest_flags("md5", &flags);
  if (err == -1) {
	fprintf(stderr, "md5 lookup failed!\n");
	return(-1);
  }

  if (digest_flag_is_enabled(flags, DIGEST_USE_ENGINE)) {
	unsigned char out[MD5_DIGEST_LEN];

	err = provider_test_md5(NULL, "abc", out);
	if (err == -1)
	  return(-1);

	fprintf(stderr, "MD5: %s\n", to_hex(out, MD5_DIGEST_LEN));
  }
  else
	fprintf(stderr, "MD5 provider test is disabled\n");

  err = crypto_get_digest_flags("sha1", &flags);
  if (err == -1) {
	fprintf(stderr, "sha1 lookup failed!\n");
	return(-1);
  }

  if (digest_flag_is_enabled(flags, DIGEST_USE_ENGINE)) {
	unsigned char out[SHA1_DIGEST_LEN];

	err = provider_test_sha1(NULL, "abc", out);
	if (err == -1)
	  return(-1);

	fprintf(stderr, "SHA1: %s\n", to_hex(out, SHA1_DIGEST_LEN));

	err = provider_test_sign(NULL);
	if (err == -1)
	  fprintf(stderr, "provider_test_sign() failed\n");
  }
  else
	fprintf(stderr, "SHA1 provider tests are disabled\n");

  return(0);
}

#else
#endif
