
/*
 * JSPR - JSon ParseR
 *
 * Simple JSON validator command.
 *
 * Quick start example:
 *   % jspr -c -pp0 < foo.json
 */

/*
 * Copyright (c) 2017-2025
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: jspr.c 3398 2025-04-09 20:30:08Z brachman $
 */

#if defined(__linux__)
#define _GNU_SOURCE
#endif

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include "jspr.h"

#ifndef MAYBE_UNUSED
#define MAYBE_UNUSED    __attribute__ ((__unused__))
#endif

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2017-2025\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: jspr.c 3398 2025-04-09 20:30:08Z brachman $";
#endif

enum {
  PROG_MAX_BUFSIZE = (100 * 1024)
};

static char buf[PROG_MAX_BUFSIZE + 1];

static inline int
streq(const char *p, const char *q)
{

    return(strcmp(p, q) == 0);
}

static inline int
strneq(const char *p, const char *q, size_t n)
{

  return(strncmp(p, q, n) == 0);
}

static int colourize_flag = 0;
static int debug_flag = 0;
static int pp_flag = 0;
static int pp_level = 0;
static int verbose_flag = 0;

#ifdef NOTDEF
/*
 * Examples when reading streamed input but not reaching EOF:
 * o int, then EOF: valid
 * o int, then decimal-point: depends on next input char
 * o int, frac: valid
 * o int, frac, 'e': depends on next char
 * o int, frac, 'e', minus/plus: depends on next char
 * o int, frac, exp: valid
 */

static Jspr_rc
parse_json_value(FILE *fp)
{
  int c;
  const char *bufp;
  size_t rem;
  Jspr_config config;
  Jspr_parser *jp;
  Jspr_rc rc;

  jspr_get_config(NULL, &config);
  config.values_limit = 4096;
  config.enable_debug = debug_flag;
  config.enable_verbose = verbose_flag;
  config.enable_colourize = colourize_flag;

  jp = NULL;
  bufp = buf;
  rem = len;
  c = 0;
  while (1) {
	const char *endp;

	if (jp == NULL)
	  jp = jspr_init_x(NULL, bufp, rem, &config);
	else
	  jspr_reset(jp, bufp, rem);

	if (multival) {
	  endp = NULL;
	  if (verbose_flag)
		fprintf(stdout, "JSON Value %d:\n", c);
	  rc = jspr_parse_value(jp, NULL, &endp);
	  c++;
	}
	else
	  rc = jspr_parse_value(jp, NULL, NULL);

	if (rc == JSPR_OK) {
	  if (verbose_flag || pp_flag) {
		jspr_print_values(stdout, jp, pp_level);
		fprintf(stdout, "OK\n\n");
	  }
	}
	else {
	  if (fp != NULL) {
		Jspr_err_state es;

		jspr_err_state(jp, &es);
		fprintf(stderr, "%s%s at line %u\n",
				(es.is_fatal ? "Fatal error: " : ""),
				jspr_strerror(rc), es.err_line);
	  }
	  break;
	}

	if (multival) {
	  rem -= (endp - bufp);
	  if (rem == 0)
		break;
	  bufp = endp;
	}
	else
	  break;
  }

  jspr_end(jp);

  return(rc);
}
#endif

/*
 * Read one complete JSON text from stdin.
 * If it is valid, optionally print it.
 * If invalid, print an error message.
 */
static Jspr_rc
parse_json_text(FILE *fp, Jspr_parser **jpp)
{
  char *p;
  size_t len, rem;
  ssize_t n;
  Jspr_config config;
  Jspr_parser *jp;
  Jspr_rc rc;

  /* Read the entire input and null terminate it. */
  p = buf;
  len = 0;
  rem = PROG_MAX_BUFSIZE;
  while (rem > 0 && (n = read(0, p, rem)) != 0) {
	p += n;
	rem -= n;
	len += n;
  }

  *p = '\0';
  if (verbose_flag > 1)
	fprintf(fp, "Read %lu bytes\n\n", len);

  if (rem == 0 && n != 0)
	return(JSPR_ERROR_INPUT_OVERFLOW);

  jspr_get_config(NULL, &config);
  config.values_limit = 4096;
  config.enable_debug = debug_flag;
  config.enable_verbose = verbose_flag;
  config.enable_colourize = colourize_flag;

  buf[len] = '\0';
  jp = jspr_init_x(NULL, buf, len, &config);
  rc = jspr_parse(jp);

  if (rc == JSPR_OK) {
	if (verbose_flag || pp_flag) {
	  jspr_print_values(fp, jp, pp_level);
	  fprintf(fp, "OK\n\n");
	}

	if (jpp != NULL)
	  *jpp = jp;
	else
	  jspr_end(jp);
  }
  else {
	Jspr_err_state es;

	jspr_err_state(jp, &es);
	fprintf(fp, "%s%s at line %u\n",
			(es.is_fatal ? "Fatal error: " : ""),
			jspr_strerror(rc), es.err_line);
	jspr_end(jp);
  }

  return(rc);
}

/***********/

static MAYBE_UNUSED Jspr_index
value_index(Jspr_parser *jp, Jspr_value *v)
{

  if (JSPR_CONFIG(jp)->enable_no_values)
    return(1);

  if (v < &jp->values[0])
    return(0);

  return((Jspr_index) (v - jp->values));
}

/*
 * Return 1 if ok, anything else causes traversal to stop with that value
 * returned.
 */
static MAYBE_UNUSED int
traverse_cb(Jspr_parser *jp, Jspr_value *v, void *arg)
{

  jspr_print_value(stdout, jp, v);

  return(1);
}

/*
 * This is a test/demo far parsing a 2fas.com JSON backup file and
 * extracting/formatting per account fields of interest.
 *
 * XXX Since I haven't found a formal schema, data extraction is based
 * on inspection of the JSON and knowledge of the app.
 */

typedef struct Tfas_info Tfas_info;
typedef struct Tfas_otp Tfas_otp;

struct Tfas_otp {
  char *algorithm;
  char *counter;
  char *digits;
  char *issuer;
  char *period;
  char *token_type;
};

struct Tfas_info {
  char *account;
  char *name;
  char *secret;
  Tfas_otp otp;
  Tfas_info *next;
};

static Tfas_info *
xstuff_parse(Jspr_parser *jp)
{
  Jspr_value *obj, *v;
  Tfas_info *tfas, *tfas_head;

  tfas_head = NULL;

  if ((v = jspr_member_lookup(jp, NULL, "services", JSPR_ARRAY_TYPE)) == NULL)
	return(NULL);

  tfas = NULL;
  for (obj = jspr_element_first(jp, v); obj != NULL;
	   obj = jspr_element_next(jp, obj)) {
	Jspr_value *el, *otp;

	if (obj->type != JSPR_OBJECT_TYPE)
	  break;

	if ((otp = jspr_member_lookup(jp, obj, "otp", JSPR_OBJECT_TYPE)) == NULL)
	  break;
	
	if (JSPR_CONFIG(jp)->enable_debug) {
	  fprintf(stdout, "-----\n");
	  jspr_breadth_first(jp, obj, traverse_cb, NULL);
	}

	tfas = (Tfas_info *) malloc(sizeof(Tfas_info));
	tfas->account = tfas->name = tfas->secret = NULL;
	tfas->otp.counter = tfas->otp.digits = tfas->otp.period = NULL;
	tfas->otp.token_type = NULL;
	tfas->next = NULL;

	if ((el = jspr_member_lookup(jp, obj, "name", JSPR_STRING_TYPE)) == NULL) {
	  free(tfas);
	  break;
	}
	tfas->name = jspr_string_copy(jp, el);

	if ((el = jspr_member_lookup(jp, obj, "secret", JSPR_STRING_TYPE)) == NULL) {
	  free(tfas->name);
	  free(tfas);
	  break;
	}
	tfas->secret = jspr_string_copy(jp, el);

	if ((el = jspr_member_lookup(jp, otp, "account", JSPR_STRING_TYPE)) != NULL)
	  tfas->account = jspr_string_copy(jp, el);

	if ((el = jspr_member_lookup(jp, otp, "algorithm", JSPR_STRING_TYPE)) != NULL)
	  tfas->otp.algorithm = jspr_string_copy(jp, el);

	if ((el = jspr_member_lookup(jp, otp, "counter", JSPR_STRING_TYPE)) != NULL)
	  tfas->otp.counter = jspr_string_copy(jp, el);

	if ((el = jspr_member_lookup(jp, otp, "digits", JSPR_NUMBER_TYPE)) != NULL)
	  tfas->otp.digits = jspr_string_copy(jp, el);

	if ((el = jspr_member_lookup(jp, otp, "issuer", JSPR_STRING_TYPE)) != NULL)
	  tfas->otp.issuer = jspr_string_copy(jp, el);

	if ((el = jspr_member_lookup(jp, otp, "period", JSPR_NUMBER_TYPE)) != NULL)
	  tfas->otp.period = jspr_string_copy(jp, el);

	if ((el = jspr_member_lookup(jp, otp, "tokenType", JSPR_STRING_TYPE)) != NULL)
	  tfas->otp.token_type = jspr_string_copy(jp, el);

	if (tfas_head == NULL) {
	  tfas_head = tfas;
	  tfas->next = NULL;
	}
	else {
	  tfas->next = tfas_head;
	  tfas_head = tfas;
	}
  }

  return(tfas_head);
}

/*
 * Parse the 2fas.com backup file, then extract and list account information.
 * If XCMD is NULL, list all services.
 * If it is "-long", list all services with detail.
 * Otherwise it is a string that is substring-matched case-insensitively against
 * each services's account and name attributes; only matching services are listed.
 */
static int
do_xstuff(Jspr_parser *jp, char *xcmd)
{
  int cnt;
  Tfas_info *tfas, *tfas_head;

  if ((tfas_head = xstuff_parse(jp)) == NULL)
	return(-1);

  cnt = 0;
  for (tfas = tfas_head; tfas != NULL; tfas = tfas->next) {
	if (xcmd != NULL && strcmp(xcmd, "-long")) {
	  if (tfas->account != NULL && strcasestr(tfas->account, xcmd) != NULL)
		;
	  else if (tfas->name != NULL && strcasestr(tfas->name, xcmd) != NULL)
		;
	  else
		continue;
	}

	fprintf(stderr, "%2d. ", ++cnt);
	fprintf(stderr, "%-15.15s -- ", tfas->name);
	if (tfas->account == NULL)
	  fprintf(stderr, "?, ");
	else		
	  fprintf(stderr, "%s, ", tfas->account);

	fprintf(stderr, "\"%s\"", tfas->secret);
	fprintf(stderr, "\n");

	if (xcmd != NULL && !strcmp(xcmd, "-long")) {
#define SEP(X)		((X != NULL) ? ", " : "")
	  fprintf(stderr, "%23.23s", " ");
	  if (tfas->otp.algorithm != NULL)
		fprintf(stderr, "%s%s", tfas->otp.algorithm, SEP(tfas->otp.digits));
	  if (tfas->otp.digits != NULL)
		fprintf(stderr, "%sd%s", tfas->otp.digits, SEP(tfas->otp.issuer));
	  if (tfas->otp.issuer != NULL)
		fprintf(stderr, "\"%s\"%s", tfas->otp.issuer, SEP(tfas->otp.period));
	  if (tfas->otp.period != NULL)
		fprintf(stderr, "%sp%s", tfas->otp.period, SEP(tfas->otp.token_type));
	  if (tfas->otp.token_type != NULL)
		fprintf(stderr, "%s", tfas->otp.token_type);
	  fprintf(stderr, "\n");
#undef SEP
	}

  }

  return(cnt);
}

/***********/

static void
usage(void)
{

  fprintf(stderr, "Usage: jspr [-d] [-h|--help] [-pp# [-c]] [-x] [-v]\n");
  fprintf(stderr, "-c:   with -pp#, colourize the JSON text.\n");
  fprintf(stderr, "-d:   debugging mode.\n");
  fprintf(stderr, "-h:   print help message, then exit.\n");
  fprintf(stderr, "-pp#: print parse information at level #.\n");
  fprintf(stderr, "-x:   Experimental: the input is a 2fas.com backup file.\n");
  fprintf(stderr, "Exit status is zero only if input was valid JSON.\n");

  exit(1);
}

int
main(int argc, char **argv)
{
  int i, xflag;
  char *xcmd;
  Jspr_parser *jp;
  Jspr_rc rc;

  colourize_flag = 0;
  pp_flag = pp_level = 0;
  xflag = 0;
  verbose_flag = 0;

  for (i = 1; i < argc; i++) {
	if (streq(argv[i], "-c"))
	  colourize_flag++;
	else if (streq(argv[i], "-d"))
	  debug_flag++;
	else if (streq(argv[i], "-h") || streq(argv[i], "--help")) {
	  usage();
	  /*NOTREACHED*/
	}
	else if (strneq(argv[i], "-pp", 3)) {
	  char *endptr;

	  pp_level = strtol(&argv[i][3], &endptr, 10);
	  if ((pp_level == 0 && errno == EINVAL) || errno == ERANGE
		  || *endptr != '\0') {
		usage();
		/*NOTREACHED*/
	  }
	  pp_flag++;
	}
	else if (streq(argv[i], "-v"))
	  verbose_flag++;
	else if (streq(argv[i], "-x")) {
	  /*
	   * If this is not the last flag, the next argument will also be grabbed.
	   * See do_xtuff().
	   */
	  xflag++;
	  xcmd = argv[++i];
	}
	else {
	  usage();
	  /*NOTREACHED*/
	}
  }

  rc = parse_json_text(stdout, &jp);
  if (rc != JSPR_OK)
	exit(1);

  if (xflag)
	do_xstuff(jp, xcmd);

  exit(0);
}
