/*
 * JSPR - JSon ParseR
 *
 * Parse tree navigation, including visiting each element, searching for
 * names, traversing an object or array, and so on.
 *
 * XXX should provide an implementation of
 * RFC 9535: JSONPath: Query Expressions for JSON
 * https://datatracker.ietf.org/doc/html/rfc9535
 */

/*
 * Copyright (c) 2017-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: jspr_nav.c 3360 2025-03-31 23:28:47Z brachman $
 */

#include <stdlib.h>
#include <stdio.h>

#include "jspr_priv.h"

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2017-2024\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: jspr_nav.c 3360 2025-03-31 23:28:47Z brachman $";
#endif

/*
 * V is an element of an Object or Array.
 * Return the "previous" element in the sibling ordering, or zero.
 *
 * Since this can be called when the parse tree is in an inconsistent state,
 * expect the unexpected and return zero if anything is amiss.
 */
Jspr_index
find_sib_prev(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_index ind, v_ind;

#ifdef SIB_PREV
  ind = v->sib_prev;
#else
  {
	Jspr_value *vp, *vs;

	ind = 0;
	if (v->parent != 0) {
	  /*
	   * Identify V and its parent, then traverse the parent's children
	   * until we find V, noting V's predecessors.
	   */
	  v_ind = value_index(jp, v);
	  vp = value_ptr(jp, v->parent);
	  /*
	   * If V is a name, it's parent is the member value... look at grandparent.
	   */
	  if (v->type == JSPR_NAME_TYPE) {
		v_ind = v->parent;
		vp = value_ptr(jp, vp->parent);
	  }

	  /* XXX Sanity checks... */
	  if (vp == NULL || vp->nchildren == 0 || vp->child_first == 0)
		return(0);

	  /* If V is the first child, it has no previous element. */
	  if ((ind = vp->child_first) == v_ind)
		return(0);

	  /* Starting with the first child, look for the value before V. */
	  vs = value_ptr(jp, ind);
	  while (vs->sib_next != v_ind && vs->sib_next != 0)
		vs = value_ptr(jp, vs->sib_next);

	  if (vs->sib_next == v_ind)
		ind = value_index(jp, vs);
	  else
		ind = 0;
	}
  }
#endif

  return(ind);
}

/*
 * PV is an Object or Array ("the parent value").
 * Return its last child element in the arbitrary ordering, or zero.
 */
Jspr_index
find_child_last(Jspr_parser *jp, Jspr_value *pv)
{
  Jspr_index ind;

  if (pv == NULL || !value_is_struct(pv))
	return(0);

#ifdef CHILD_LAST
  ind = pv->child_last;
#else
  {
	Jspr_value *el;

	if ((ind = pv->child_first) == 0)
	  return(0);
	if ((el = value_ptr(jp, ind)) == NULL)
	  return(0);

	while (el->sib_next != 0) {
	  ind = el->sib_next;
	  if ((el = value_ptr(jp, ind)) == 0)
		return(0);
	};
  }
#endif

  return(ind);
}

/*
 * If ROOT is an Object, return the number of (first-level) members;
 * if ROOT is an Array, return the number of array elements/values;
 * otherwise, return -1.
 * If ROOT is NULL, use the top-most value, which must be an Object or Array.
 * Note than an Object or Array can have zero elements.
 */
int
jspr_element_count(Jspr_parser *jp, Jspr_value *root)
{

  if (root == NULL)
	root = &jp->values[JSPR_VALUE_ROOT_INDEX];

  if (!value_is_struct(root))
	return(-1);

  return(root->nchildren);
}

#ifdef NOTDEF
/*
 * Return the total number of siblings of V, including V (i.e., the
 * number of children that the parent of V has).
 */
int
jspr_siblings_count(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_value *vp;

  if (v == NULL || v->parent == 0)
	return(0);

  vp = value_ptr(jp, v->parent);

  if (!value_is_struct(vp))
	return(0);

  return(vp->nchildren);
}
#endif

/*
 * Return the number of children of VALUE, recursively.
 * Only an Array or Object can have children.
 */
int
jspr_child_count(Jspr_parser *jp, Jspr_value *value)
{
  int total;
  Jspr_index ind;
  Jspr_value *v;

  if (!value_is_struct(value))
    return(-1);

  total = value->nchildren;
  ind = value->child_first;
  while (ind < JSPR_CONFIG(jp)->num_values) {
	v = value_ptr(jp, ind);
	if (value_is_struct(v)) {
	  if (v->nchildren) {
		total += v->nchildren;
		total += jspr_child_count(jp, v);
	  }
	}
	ind = v->sib_next;
	if (ind == 0)
	  break;
  }

  return(total);
}

/*
 * If ROOT is an Object or Array, return the NTH member/element (starting
 * at zero) if it exists, otherwise the NULL pointer.
 * XXX not particularly efficient if there are a lot of elements
 */
Jspr_value *
jspr_element_nth(Jspr_parser *jp, Jspr_value *root, unsigned int nth)
{
  int c;
  Jspr_index ind;
  Jspr_value *start, *v;

  if (root == NULL)
	start = &jp->values[JSPR_VALUE_ROOT_INDEX];
  else {
	if (value_is_struct(root))
	  start = root;
	else if (root->parent != 0) {
	  start = value_ptr(jp, root->parent);
	  if (!value_is_struct(start))
		return(NULL);
	}
	else
	  return(NULL);
  }

  if (start->nchildren == 0 || nth > start->nchildren)
	return(NULL);

  c = 0;
  ind = start->child_first;
  while (ind != 0) {
	v = value_ptr(jp, ind);
	if (c == nth)
	  return(v);
	ind = v->sib_next;
	if (ind == 0)
	  break;
	c++;
  }

  return(NULL);
}

/*
 * Return the parent value of V, or NULL.
 */
Jspr_value *
jspr_value_parent(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_value *pv;

  if (v == NULL || v->parent == 0)
	return(NULL);

  pv = value_ptr(jp, v->parent);

  return(pv);
}

/*
 * V (which defaults to the root value) must be an Object or Array.
 * Return its first element in the arbitrary ordering, or NULL if it is empty.
 */
Jspr_value *
jspr_element_first(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_value *el, *pv;

  if ((pv = v) == NULL)
	pv = &jp->values[JSPR_VALUE_ROOT_INDEX];

  if (!value_is_struct(pv))
	return(NULL);

  if (pv->nchildren == 0)
	return(NULL);

  el = value_ptr(jp, pv->child_first);
  return(el);
}

/*
 * V (which defaults to the root value) must be an Object or Array.
 * Return its last element, or NULL if it is empty.
 */
Jspr_value *
jspr_element_last(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_index ind;
  Jspr_value *el, *pv;

  if ((pv = v) == NULL)
	pv = &jp->values[JSPR_VALUE_ROOT_INDEX];

  if (!value_is_struct(pv))
	return(NULL);

  if (pv->nchildren == 0)
	return(NULL);

  if ((ind = find_child_last(jp, pv)) == 0)
	return(NULL);

  el = value_ptr(jp, ind);
  return(el);
}

/*
 * V must be an element of an Object or Array.
 * Return the next element, or NULL if there are no more elements.
 * The ordering of elements is fixed but indeterminate.
 */
Jspr_value *
jspr_element_next(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_value *next_v;

  if (v == NULL || v->sib_next == 0)
	return(NULL);

  next_v = value_ptr(jp, v->sib_next);
  return(next_v);
}

/*
 * V must be an element of an Object or Array.
 * Return the previous element, or NULL if there are no more elements.
 * The ordering of elements is fixed but indeterminate.
 */
Jspr_value *
jspr_element_previous(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_value *prev_v;

#ifdef SIB_PREV
  if (v == NULL || v->sib_prev == 0)
	return(NULL);

  prev_v = value_ptr(jp, v->sib_prev);
#else
  prev_v = NULL;
#endif
  return(prev_v);
}

Jspr_value *
jspr_value_root(Jspr_parser *jp)
{

  if (jp == NULL || jp->next_value == JSPR_VALUE_ROOT_INDEX)
	return(NULL);

  return(&jp->values[JSPR_VALUE_ROOT_INDEX]);
}

/*
 * Return the value that was created immediately before V, if possible.
 */
Jspr_value *
jspr_value_previous(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_index ind;
  Jspr_value *pv;

  if (v == NULL)
	return(NULL);

  /* Check if V is an invalid value or if there is no previous value. */
  if ((ind = value_index(jp, v)) == 0 || ind == JSPR_VALUE_ROOT_INDEX)
	return(NULL);

  pv = &jp->values[ind - 1];

  return(pv);
}

/*
 * Return the value that was created immediately after V, if possible.
 */
Jspr_value *
jspr_value_next(Jspr_parser *jp, Jspr_value *v)
{
  Jspr_index ind;
  Jspr_value *nv;

  if (v == NULL)
	return(NULL);

  /* Check if V is an invalid value or if there is no next value. */
  if ((ind = value_index(jp, v)) == 0 || ind >= (jp->next_value - 1))
	return(NULL);

  nv = &jp->values[ind + 1];

  return(nv);
}

static int debug_value_name_eq = 0;

static int
value_name_eq(Jspr_parser *jp, Jspr_value *value, char *name, size_t name_len)
{
  int rc;

  if ((js_end(jp, value) - js_start(jp, value)) == name_len
	  && *js_end(jp, value) == '"') {
	rc = strneq(name, js_start(jp, value), name_len);
	if (debug_value_name_eq) {
	  char *s;

	  s = jspr_string_copy(jp, value);
	  fprintf(stderr, "value_name_eq: \"%s\" : \"%s\" = %d\n", name, s, rc);
	  free(s);
	}

	return(rc);
  }

  return(0);
}

typedef struct Cb_member_lookup {
  char *name;
  size_t name_len;
  Jspr_type type;
  Jspr_value *result;
} Cb_member_lookup;

/*
 * Return 0 if V matches ARG and the lookup should terminate, otherwise
 * return 1 to continue the search.
 */
static int
cb_member_lookup(Jspr_parser *jp, Jspr_value *v, void *arg)
{
  Cb_member_lookup *m;
  Jspr_value *vn, *vp;

  m = (Cb_member_lookup *) arg;
  if (v->parent != 0) {
	vp = value_ptr(jp, v->parent);
	if (vp->type == JSPR_OBJECT_TYPE && v->name != 0) {
	  vn = value_ptr(jp, v->name);
	  if (JSPR_CONFIG(jp)->enable_debug) {
		char *s;

		s = jspr_string_copy(jp, vn);
		fprintf(stderr, "cb_member_lookup: \"%s\"\n", s);
		free(s);
	  }

	  if (m->type != JSPR_ANY_TYPE
		  && m->type != JSPR_STRING_TYPE
		  && v->type != m->type)
		return(1);

	  /* Found - set the result and terminate the search. */
	  if (value_name_eq(jp, vn, m->name, m->name_len) == 1) {
		m->result = v;
		return(0);
	  }
	}
  }

  return(1);
}

/*
 * Return the first instance of an Object member named NAME found using
 * a breadth-first search beginning at value node V (NULL implies the root
 * node of the parse tree) that is of TYPE (ignored if JSPR_ANY_TYPE).
 * Return NULL if it is not found.
 *
 * XXX this could be extended to regex match the name
 */
Jspr_value *
jspr_member_lookup_x(Jspr_parser *jp, Jspr_lookup_alg alg, Jspr_value *v,
					 char *name, Jspr_type type)
{
  Cb_member_lookup m;

  m.name = name;
  m.name_len = strlen(name);
  m.type = type;
  m.result = NULL;

  if (alg == JSPR_LOOKUP_BREADTH_FIRST) {
	if (jspr_breadth_first(jp, v, cb_member_lookup, &m) == 0)
	  return(m.result);
  }
  else if (alg == JSPR_LOOKUP_PREORDER) {
	if (jspr_preorder(jp, v, cb_member_lookup, &m) == 0)
	  return(m.result);
  }

  return(NULL);
}

Jspr_value *
jspr_member_lookup(Jspr_parser *jp, Jspr_value *v, char *name, Jspr_type type)
{
  Jspr_value *r;

  r = jspr_member_lookup_x(jp, JSPR_LOOKUP_BREADTH_FIRST, v, name, type);

  return(r);
}

/*
 * Starting at ROOT, return the "first" member value that has name NAME
 * and is of type TYPE, or NULL if one is not found.
 * The default is to start at the first value node.
 * The special type JSPR_ANY_TYPE matches all types.
 *
 * The search is done for efficiency - if there are multiple instances
 * of the same name any one of them might be returned.
 * RFC 8259 does not require unique names anywhere.
 */
Jspr_value *
jspr_member_lookup_fast(Jspr_parser *jp, Jspr_value *root, char *name,
						Jspr_type type)
{
  int i;
  size_t name_len;
  Jspr_index start_ind;
  Jspr_value *v;

  if (root == NULL)
	start_ind = JSPR_VALUE_ROOT_INDEX;
  else
	start_ind = value_index(jp, root);

  name_len = strlen(name);
  for (i = start_ind; i < jp->next_value; i++) {
	v = &jp->values[i];
	if (v->type == JSPR_NAME_TYPE) {
	  Jspr_value *vn;

	  vn = value_ptr(jp, v->parent);
	  if (type != JSPR_ANY_TYPE && vn->type != type)
		continue;

	  if (jspr_value_name_eq(jp, v, name, name_len))
		return(value_ptr(jp, v->parent));
	}
  }

  return(NULL);
}

/*
 * Preorder traversal of the parse tree starting at value node ROOT
 * (if NULL, then the root node of the parse tree).
 *
 * Call CB with the value and user-supplied argument, exactly once for each
 * value. Traversal will continue if the callback function returns 1,
 * otherwise it stops and returns that value.
 *
 * Return 1 if the search was unsuccessful, otherwise return whatever
 * value (not 1) that CB returned.
 */
int
jspr_preorder(Jspr_parser *jp, Jspr_value *root,
			  int (*cb)(Jspr_parser *jp, Jspr_value *v, void *arg), void *arg)
{
  int rc;
  Jspr_index v_ind;
  Jspr_value *start, *vc, *vs;

  if ((start = root) == NULL)
	start = &jp->values[JSPR_VALUE_ROOT_INDEX];

  vc = start;
  while ((v_ind = vc->child_first) != 0) {
	vc = value_ptr(jp, v_ind);
	if ((rc = jspr_preorder(jp, vc, cb, arg)) != 1)
	  return(rc);
  }

  if ((rc = cb(jp, start, arg)) != 1)
	return(rc);

  v_ind = start->sib_next;
  while (v_ind != 0) {
	vs = value_ptr(jp, v_ind);
	if (vs->nchildren) {
	  if ((rc = jspr_preorder(jp, value_ptr(jp, vs->child_first), cb, arg)) != 1)
		return(rc);
	}

	if ((rc = cb(jp, vs, arg)) != 1)
	  return(rc);
	v_ind = vs->sib_next;
  }

  return(1);
}

/*
 * Search starting with the immediate children of ROOT (default is the root of
 * the parse tree), from left-to-right, then recurse into each child node
 * (left-to-right); e.g.,
 *                  Root
 *             /     /     /
 *            1     2     3
 *         / / /    / /   / /  /
 *        4 5 6    7 8   9 10 11
 *
 * Call CB with the value and user-supplied argument, exactly once for each
 * value. Traversal will continue if the callback function returns 1,
 * otherwise it stops and returns that value.
 *
 * Return 1 if the search was unsuccessful, otherwise return whatever
 * value (not 1) that CB returned.
 */
int
jspr_breadth_first(Jspr_parser *jp, Jspr_value *root,
				   int (*cb)(Jspr_parser *jp, Jspr_value *v, void *arg),
				   void *arg)
{
  int rc;
  Jspr_index v_ind;
  Jspr_value *start, *vc, *vs;

  if ((start = root) == NULL)
	start = &jp->values[JSPR_VALUE_ROOT_INDEX];

  v_ind = value_index(jp, start);
  if (v_ind != 0) {
	Jspr_index child;

	/* Visit each of the starting node's children, left-to-right. */
	vs = value_ptr(jp, v_ind);
	child = vs->child_first;
	while (child != 0) {
	  vc = value_ptr(jp, child);
	  if ((rc = cb(jp, vc, arg)) != 1)
		return(rc);
	  child = vc->sib_next;
	}

	/* Recurse into each of the starting node's children, left-to-right. */
	child = vs->child_first;
	while (child != 0) {
	  vc = value_ptr(jp, child);
	  if ((rc = jspr_breadth_first(jp, vc, cb, arg)) != 1)
		return(rc);
	  child = vc->sib_next;
	}
  }

  return(1);
}
