/*
 * Copyright (c) 2003-2004 DSS Distributed Systems Software, Inc.
 * All rights reserved.   See the file LICENSE for redistribution information.
 */

/*****************************************************************************
 * COPYRIGHT AND PERMISSION NOTICE
 * 
 * Copyright (c) 2001-2003 The Queen in Right of Canada
 * 
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, and/or sell
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, provided that the above copyright notice(s) and this
 * permission notice appear in all copies of the Software and that both the
 * above copyright notice(s) and this permission notice appear in supporting
 * documentation.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE 
 * BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 * 
 * Except as contained in this notice, the name of a copyright holder shall not
 * be used in advertising or otherwise to promote the sale, use or other
 * dealings in this Software without prior written authorization of the
 * copyright holder.
 ***************************************************************************/

/*
 * mkargv - scan a string and create an argv vector
 *
 * Given STR, find its individual elements and assign each to its relative
 * position in the argv vector.  Each element is copied using malloc(3).
 * The argv vector of pointers is also allocated using malloc(3).
 * The input string is terminated (eos) by a newline or null character.
 *
 * Examples:
 * The call: mkargv("hello world!", NULL, &argv);
 * would return 2, with argv[0] set to "hello" (no quotes), argv[1] set to
 * "world!" (no quotes), and argv[2] set to NULL.
 *
 * The call: mkargv("{Auggie's string}", NULL, &argv);
 * would return 1, with argv[0] set to "Auggie's string" (no quotes)
 * and argv[1] set to NULL.
 *
 * Each element is demarcated as follows:
 * o initial whitespace (blanks or tabs) is skipped
 * o Optionally, a starting quote character signals the start of the element
 * o The element continues until eos or whitespace is found (the no quote case)
 *   or the matching ending quote character is found followed by eos or
 *   whitespace
 * o Any character other than eos can be quoted by being preceded with
 *   a backslash character
 *
 * Different quote characters may be used to demarcate different elements.
 *
 * If KEEPQ is non-zero, any starting and ending quote characters are
 * kept in the copied elements.
 *
 * The pointer to the last element is followed by a NULL pointer.
 *
 * Errors are indicated by the following values:
 *   -1: malloc failed
 *   -2: a matching closing quote is missing
 *   -3: STR ends with a backslash character
 * 
 * The number of elements in STR is returned, or a negative value (as above)
 * on error.
 */

#include "mkargv.h"

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2003-2004\nDSS Distributed Systems Software, Inc.\n\
All rights reserved.\n";
static MAYBE_UNUSED const char revid[] =
  "$Id: mkargv.c 3352 2024-09-23 22:41:06Z brachman $";
#endif

static const char *default_startq = "\"'{";
static const char *default_endq   = "\"'}";
static const char *default_ifs = " \t";
static const int default_keepq = 0;
static const int default_keepws = 0;

typedef struct Argv {
  char **argv;
  int nalloc;
  int nused;
  int incr;
} Argv;

static Argv *
expand(Argv *a)
{

  a->argv = (char **) realloc(a->argv, sizeof(char *) * (a->nalloc + a->incr));
  if (a->argv == NULL)
	return(NULL);

  a->nalloc += a->incr;
  return(a);
}

static Argv *
init(Argv *a, int n)
{

  a->argv = NULL;
  if (n) {
    a->argv = (char **) malloc(sizeof(char *) * n);
    if (a->argv == NULL) {
	  free(a);
	  return(NULL);
    }
  }

  a->nalloc = n;
  a->nused = 0;
  a->incr = n;
  return(a);
}

static char *
copyarg(char *str, int len)
{
  char *c, *e, *p, *s;

  if ((p = c = (char *) malloc(len + 1)) == NULL)
	return(NULL);

  s = str;
  e = s + len;
  while (s < e) {
	if (*s == '\\')
	  s++;
	*p++ = *s++;
  }

  *p = '\0';
  return(c);
}

static int
add(Argv *argv, char *str, int len)
{
  char *w;

  if (argv->incr != 0) {
	if (argv->nalloc == argv->nused) {
	  if (expand(argv) == NULL)
		return(-1);
	}
	w = NULL;
	if (str != NULL) {
	  if ((w = copyarg(str, len)) == NULL)
		return(-1);
	}
	argv->argv[argv->nused++] = w;
  }
  else
    argv->nused++;

  return(0);
}

int
mkargv(char *str, Mkargv *conf, char ***argvp)
{
  int endq_ch, err, keepq, keepws;
  char *p, *q, *s;
  const char *ifs, *startq, *endq;
  Argv argv;

  if (conf != NULL) {
	keepq = conf->keepq;
	keepws = conf->keepws;
	if (conf->ifs != NULL)
	  ifs = conf->ifs;
	else
	  ifs = default_ifs;
	if (conf->startq != NULL && conf->endq != NULL) {
	  startq = conf->startq;
	  endq = conf->endq;
	}
	else {
	  startq = default_startq;
	  endq = default_endq;
	}
  }
  else {
	keepq = default_keepq;
	keepws = default_keepws;
	ifs = default_ifs;
	startq = default_startq;
	endq = default_endq;
  }

  p = str;
  err = -1;

  if (init(&argv, (argvp == NULL) ? 0 : 8) == NULL)
	return(err);

  while (1) {
	if (!keepws) {
	  /* Skip leading white space */
 	  while (*p == ' ' || *p == '\t')
	    p++;
	}

	if (*p == '\n' || *p == '\0')
	  goto done;

	/* Note that start */
	s = p;

	/* If there's an optional quote, note that we need a matching end quote */
	endq_ch = 0;
	if ((q = strchr(startq, (int) *p)) != NULL) {
	  endq_ch = endq[q - startq];
	  if (!keepq)
		s = p + 1;
	  p++;
	}

	/* Find the end of the argument */
	while (1) {
	  if (*p == '\n' || *p == '\0') {
		if (endq_ch) {
		  /* Missing closing quote */
		  err = -2;
		  goto fail;
		}

		if (add(&argv, s, p - s) == -1) {
		  err = -1;
		  goto fail;
		}
		break;
	  }
	  else if (endq_ch && *p == endq_ch
			   && (strchr(ifs, (int) *(p + 1)) != NULL || *(p + 1) == '\n'
				   || *(p + 1) == '\0')) {
		if (add(&argv, s, keepq ? (p - s + 1) : (p - s)) == -1) {
		  err = -1;
		  goto fail;
		}
		p++;
		break;
	  }
	  else if (*p == '\\') {
		p++;
		if (*p == '\n' || *p == '\0') {
		  /* Trailing escape character */
		  err = -3;
		  goto fail;
		}
		p++;
	  }
	  else if (!endq_ch && (strchr(ifs, (int) *p) != NULL)) {
		if (add(&argv, s, p - s) == -1) {
		  err = -1;
		  goto fail;
		}
		p++;
		break;
	  }
	  else
		p++;
	}
  }

 done:
  add(&argv, NULL, 0);
  if (argvp != NULL)
    *argvp = argv.argv;
  return(argv.nused - 1);

 fail:
  free(argv.argv);
  return(err);
}

#ifdef PROG
int
main(int argc, char **argv)
{
  int i, n;
  char **a, *ifs, buf[1024];
  Mkargv conf;
  
  ifs = NULL;
  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-ifs") == 0) {
      if (argv[++i] != NULL)
        ifs = argv[i];
      else {
        fprintf(stderr, "?\n");
        exit(1);
      }
    }
    else {
      fprintf(stderr, "?\n");
      exit(1);
    }
  }

  conf.keepq = 0;
  conf.keepws = 0;
  conf.ifs = ifs;
  conf.startq = NULL;
  conf.endq = NULL;

  while (fgets(buf, sizeof(buf), stdin) != NULL) {
	if ((n = mkargv(buf, &conf, &a)) == -1)
	  fprintf(stderr, "Error\n");
	else {
	  for (i = 0; i < n; i++)
		fprintf(stderr, "%d. %s\n", i, a[i]);
	}
  }

  exit(0);
}
#endif
