/*
 * Copyright (c) 2003-2023
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * local_ldap_auth - LDAP/Microsoft Active Directory authentication module
 * local_ldap_roles - LDAP/Microsoft Active Directory roles module
 *
 * Each entry in an LDAP directory has a unique distinguished name (DN).
 * A user is authenticated if we can take the given username, form a DN,
 * and use the given password to bind to that DN.
 *
 * If authentication succeeds, then optionally do a search operation to
 * return attributes required to construct roles information.
 *
 * In addition to the username provided during login, we need configuration
 * information to allow us to construct the DN of the entry for the user.
 * In some cases, we can directly produce a DN given the username by
 * making a template-based substitution.

 * In other cases this isn't possible; we may need to first bind to a
 * specified DN (possibly using a specified password), do a search for the
 * (presumably unique) attribute having the given username as its value
 * (guided by a DN at which to base the search and a filter produced by
 * a template-based substitution.
 * The result of a successful search provides the DN for the user's entry.
 * For example, a directory schema might associate an email address with
 * each user's entry but email addresses might not be used to name entries.
 * We might bind to the directory as the administrator,
 * perform a search for the entry containing the email address, then
 * bind to the DN of that entry.
 *
 * Here are the two methods:
 * 1) If you know how to map the USERNAME into the corresponding DN, create the
 *    DN and bind to it using PASSWORD.  If the bind works, the user is
 *    authenticated.
 *
 * 2) If you don't know how to map the USERNAME directly into the corresponding
 *    DN, bind to the directory as an administrator (we're given the DN and
 *    password to use) and do a search for USERNAME (with some
 *    configuration guidance).  If the search returns a single result, we
 *    use its DN and the PASSWORD to try to bind to the directory.
 *
 * In either situation, the DACS_USERNAME could optionally be obtained from
 * an attribute value or by mapping USERNAME.
 * The directive
 *    AUTH_LDAP_MAP_USERNAME="map [-r string regex ] attribute-name [flag]*"
 * causes DACS_USERNAME to be set to the value of the specified attribute
 * name, if it's present in the entry.
 *
 * If we follow mod_auth_ldap's approach (AuthLDAPGroupAttribute Directive),
 * we'd simply concatenate the values of the selected attributes.
 *
 * For ideas, see:
 * http://httpd.apache.org/docs-2.0/mod/mod_auth_ldap.html
 * http://www.rfc-editor.org/rfc/rfc2254.txt
 * http://www.rfc-editor.org/rfc/rfc2255.txt
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2023\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: local_ldap_auth.c 3229 2023-02-07 20:40:10Z brachman $";
#endif

#include "dacs.h"

#include <ldap.h>
#include <dacs_ldif.h>

enum {
  LDAP_DEFAULT_PORT     = 389,
  LDAP_DEFAULT_SSL_PORT = 636
};

/* An attribute-value assertion */
typedef struct Ava {
  char *attr_name;
  char *attr_value;
} Ava;

static char *log_module_name = "local_ldap_auth";

static int local_ldap_roles = 0;

extern int ldap_bind_s(LDAP *, LDAP_CONST char *, LDAP_CONST char *, int);
extern LDAP *ldap_init(LDAP_CONST char *, int);
extern int ldap_unbind(LDAP *);

static int
do_bind(LDAP *ld, char *binddn, char *password)
{
  int rc;

  if (password == NULL || *password == '\0')
    log_msg((LOG_DEBUG_LEVEL, "Note: password is NULL or empty"));

  rc = ldap_bind_s(ld, binddn, password, LDAP_AUTH_SIMPLE);
  if (rc != LDAP_SUCCESS) {
	log_msg((LOG_ERROR_LEVEL, "ldap_bind_s as \"%s\" failed", binddn));
	log_msg((LOG_ERROR_LEVEL, "OpenLDAP error %d: %s", rc, ldap_err2string(rc)));
	return(-1);
  }
  log_msg((LOG_DEBUG_LEVEL, "ldap_bind_s as \"%s\" succeeded", binddn));

  return(0);
}

static char *
make_ldif(int type, char *name, char *value, ber_len_t vallen)
{
  char *ldif;

  if ((ldif = ldif_put(type, name, value, vallen)) == NULL)
	return(NULL);

  return(ldif);
}

static int
print_entry(LDAP *ld, char *dn, LDAPMessage *entry, int attrsonly)
{
  int i, rc;
  char *a;
  BerElement *ber = NULL;
  struct berval **bvals;
  LDAPControl **ctrls = NULL;

  log_msg((LOG_TRACE_LEVEL, "Entry: \"%s\"", dn));

  rc = ldap_get_entry_controls(ld, entry, &ctrls);

  if (rc != LDAP_SUCCESS) {
	log_msg((LOG_ERROR_LEVEL, "ldap_get_entry_controls: error=\"%s\", rc = %d",
			 ldap_err2string(rc), rc));
	return(-1);
  }

  for (a = ldap_first_attribute(ld, entry, &ber); a != NULL;
	   a = ldap_next_attribute(ld, entry, ber)) {
	if (attrsonly)
	  log_msg((LOG_TRACE_LEVEL, "%s",
			   strtrim(make_ldif(LDIF_PUT_NOVALUE, a, NULL, 0), "\n", 0)));
	else if ((bvals = ldap_get_values_len(ld, entry, a)) != NULL) {
	  for (i = 0; bvals[i] != NULL; i++) {
		log_msg((LOG_TRACE_LEVEL, "%s",
				 strtrim(make_ldif(LDIF_PUT_VALUE, a,
								   bvals[i]->bv_val, bvals[i]->bv_len),
						 "\n", 0)));
	  }
	}
  }

  return(0);
}

static int
print_result(LDAP *ld, LDAPMessage *result, int search)
{
  int err, rc;
  char *matched_dn, **refs, *text;
  LDAPControl **ctrls;

  matched_dn = NULL;
  text = NULL;
  refs = NULL;
  ctrls = NULL;

  rc = ldap_parse_result(ld, result,
						 &err, &matched_dn, &text, &refs, &ctrls, 0);

  if (rc != LDAP_SUCCESS) {
	log_msg((LOG_ERROR_LEVEL, "ldap_parse_result: error=\"%s\", rc=%d",
			 ldap_err2string(rc), rc));
	return(-1);
  }

  log_msg((LOG_TRACE_LEVEL, "ldap_parse_result: result=%d %s",
		   err, ldap_err2string(err)));

  if (matched_dn && matched_dn[0] != '\0')
	log_msg((LOG_TRACE_LEVEL, "Matched DN: %s", matched_dn));

  if (text && text[0] != '\0')
	log_msg((LOG_TRACE_LEVEL, "Additional information: %s", text));

  if (refs) {
	int i;

	for (i = 0; refs[i] != NULL; i++)
	  log_msg((LOG_TRACE_LEVEL, "Referral: %s", refs[i]));
  }

  return(err);
}

#ifdef NOTDEF
/*
 * Count the number of attribute/value pairs in the entry.
 * Return that number or -1 on error.
 */
static int
count_avas(LDAP *ld, LDAPMessage *entry)
{
  int i, n;
  char *a;
  BerElement *ber = NULL;
  struct berval **bvals;

  n = 0;
  for (a = ldap_first_attribute(ld, entry, &ber); a != NULL;
	   a = ldap_next_attribute(ld, entry, ber)) {
	if ((bvals = ldap_get_values_len(ld, entry, a)) != NULL) {
	  for (i = 0; bvals[i] != NULL; i++)
		;
	  n += i;
	}
  }

  return(n);
}
#endif

/*
 * Extract all of the attribute-value pairs in ENTRY and convert them
 * into a vector of AVA structures.
 * Return the number of AVA structures or -1 on error.
 * An extra AVA structure is added at the end and has NULL fields.
 */
static int
get_entry_avas(LDAP *ld, LDAPMessage *entry, Dsvec *avas)
{
  int anum, i;
  char *attr, *aname, *avalue, *ava, *p;
  BerElement *ber = NULL;
  Ava *a;
  struct berval **bvals;

#ifdef NOTDEF
  n = count_avas(ld, entry);
  if (n <= 0)
	return(n);
#endif

  anum = 0;
  for (attr = ldap_first_attribute(ld, entry, &ber); attr != NULL;
	   attr = ldap_next_attribute(ld, entry, ber)) {
	if ((bvals = ldap_get_values_len(ld, entry, attr)) != NULL) {
	  for (i = 0; bvals[i] != NULL; i++) {
		/*
		 * Unfortunately, it doesn't appear as though we're provided
		 * with a nice library function to get at the attribute value
		 * as a simple character string.
		 * So we whip something up by using a library function and
		 * then breaking it into pieces; it gives us the attribute
		 * name, followed by a colon and a space, and then the
		 * attribute value.
		 */
		ava = make_ldif(LDIF_PUT_VALUE, attr,
						bvals[i]->bv_val, bvals[i]->bv_len);
		aname = ava;
		if ((p = strchr(ava, ':')) == NULL)
		  return(-1);
		/* Multiple colons may appear... */
		while (*(p + 1) == ':')
		  p++;
		*p++ = '\0';
		if (*p++ != ' ')
		  return(-1);
		avalue = p;
		if ((p = strchr(avalue, '\n')) == NULL)
		  return(-1);
		*p = '\0';

		log_msg((LOG_TRACE_LEVEL,
				 "attr_name=\"%s\", attr_value=\"%s\"", aname, avalue));

		a = ALLOC(Ava);
		a->attr_name = aname;
		a->attr_value = avalue;
		dsvec_add_ptr(avas, a);
		anum++;
	  }
	}
  }

  return(anum);
}

/*
 * Use a Search operation to locate a particular entry.
 */
static int
fetch_entry(LDAP *ld, char *base, int scope, char *filtpatt, char *value,
			char **attrs, int attrsonly, LDAPControl **sctrls,
			LDAPControl **cctrls, struct timeval *timeout,
			int sizelimit, char **dnp, Dsvec *avas)
{
  int nentries, navas, rc;
  char *dn, *filter;
  LDAPMessage *res, *msg;
  ber_int_t msgid;

  if (filtpatt != NULL)
	filter = ds_xprintf(filtpatt, value);
  else
	filter = value;

  log_msg((LOG_TRACE_LEVEL, "Searching: base=\"%s\", filter=\"%s\"",
		   base, filter));

  rc = ldap_search_ext(ld, base, scope, filter, attrs, attrsonly,
					   sctrls, cctrls, timeout, sizelimit, &msgid);

  if (rc != LDAP_SUCCESS) {
	log_msg((LOG_ERROR_LEVEL,
			 "ldap_search_ext failed: %s (%d)", ldap_err2string(rc), rc));
	return(-1);
  }

  nentries = 0;
  navas = 0;
  res = NULL;

  while ((rc = ldap_result(ld, LDAP_RES_ANY, LDAP_MSG_ONE, NULL, &res)) > 0) {
	for (msg = ldap_first_message(ld, res); msg != NULL;
		 msg = ldap_next_message(ld, msg)) {
	  dn = ldap_get_dn(ld, msg);

	  switch (ldap_msgtype(msg)) {
	  case LDAP_RES_SEARCH_ENTRY:
		if (nentries++ == 0) {
		  if ((navas = get_entry_avas(ld, msg, avas)) == -1)
			log_msg((LOG_ERROR_LEVEL, "get_entry_avas failed"));
		  log_msg((LOG_TRACE_LEVEL, "Got %d avas", navas));
		  if (dnp != NULL)
			*dnp = dn;
		}
		log_msg((LOG_TRACE_LEVEL, "LDAP_RES_SEARCH_ENTRY:%s", dn));
		print_entry(ld, dn, msg, attrsonly);
		break;

	  case LDAP_RES_SEARCH_RESULT:
		log_msg((LOG_TRACE_LEVEL, "LDAP_RES_SEARCH_RESULT:%s", dn));
		print_result(ld, msg, 1);
		goto done;

	  default:
	  case LDAP_RES_SEARCH_REFERENCE:
	  case LDAP_RES_EXTENDED:
#ifdef LDAP_RES_EXTENDED_PARTIAL
	  case LDAP_RES_EXTENDED_PARTIAL:
#endif
#ifdef LDAP_RES_INTERMEDIATE
	  case LDAP_RES_INTERMEDIATE:
#endif
		goto done;
	  }
	}

	ldap_msgfree(res);
  }

  if (rc == -1)
	log_msg((LOG_ERROR_LEVEL, "ldap_result: %s", ldap_err2string(rc)));

 done:

  if (rc == -1)
	return(-1);

  return(nentries);
}

/*
 * Define the string prefix common to all OpenLDAP's option names.
 * In an attempt to reduce confusion, we will use the same option names as
 * OpenLDAP uses (see ldap_get_option(3)).
 *
 * To set an OpenLDAP option, use an OPTION directive within the <Auth> clause.
 * Only OPTION directive variable names that begin with this prefix will be
 * examined; for instance:
 *   OPTION 'LDAP_OPT_X_TLS=LDAP_OPT_X_TLS_HARD'
 *   OPTION 'LDAP_OPT_DEBUG_LEVEL=1'
 *   OPTION 'LDAP_OPT_X_TLS_CACERTFILE=/path/to/server/cert/server.crt'
 * To quickly comment out one of these options, just use a different prefix:
 *   OPTION 'XXXLDAP_OPT_DEBUG_LEVEL=1'
 *
 * XXX There is currently no way to use an OPTION directive to remove
 * an OpenLDAP option from the environment.  Consider introducing a second
 * prefix (e.g., "LDAP_OPT_REMOVE_") that would remove the corresponding
 * environment variable before OpenLDAP sees it.
 */
#define LDAP_OPTION_PREFIX	"LDAP_OPT_"

/* The four flavours of option supported here. */
typedef enum Ldap_option_type {
  LDAP_OPT_BOOL = 0,
  LDAP_OPT_INT  = 1,
  LDAP_OPT_STR  = 2,
  LDAP_OPT_ENUM = 3,
  LDAP_OPT_NONE = 4
} Ldap_option_type;

/* Symbolic values defined by OpenLDAP that we want to recognize. */
typedef struct Ldap_option_enum {
  char *name;			/* OpenLDAP's symbolic name. */
  int intval;			/* Integer value of that symbol. */
} Ldap_option_enum;

/*
 * Mappings between symbols defined by OpenLDAP and their corresponding
 * integer values.
 * 
 * The enum values (ints) in this table may not be unique.
 * Refer to ldap_ set_option(3) and ldap.conf(5).
 * Also: OpenLDAP's ldap.h, options.c, init.c
 * Also: https://www.openldap.org/doc/admin21/tls.html
 *
 * XXX This list is incomplete; extend as necessary.
 */
static Ldap_option_enum ldap_option_enums[] = {
  { "LDAP_OPT_X_TLS_NEVER",           LDAP_OPT_X_TLS_NEVER },
  { "LDAP_OPT_X_TLS_HARD",            LDAP_OPT_X_TLS_HARD },
  { "LDAP_OPT_X_TLS_DEMAND",          LDAP_OPT_X_TLS_DEMAND },
  { "LDAP_OPT_X_TLS_ALLOW",           LDAP_OPT_X_TLS_ALLOW },
  { "LDAP_OPT_X_TLS_TRY",             LDAP_OPT_X_TLS_TRY },
  { "LDAP_OPT_X_TLS_CRL_NONE",        LDAP_OPT_X_TLS_CRL_NONE },
  { "LDAP_OPT_X_TLS_CRL_PEER",        LDAP_OPT_X_TLS_CRL_PEER },
  { "LDAP_OPT_X_TLS_CRL_ALL",         LDAP_OPT_X_TLS_CRL_ALL },
  { "LDAP_OPT_X_TLS_PROTOCOL_SSL2",   LDAP_OPT_X_TLS_PROTOCOL_SSL2 },
  { "LDAP_OPT_X_TLS_PROTOCOL_SSL3",   LDAP_OPT_X_TLS_PROTOCOL_SSL3 },
  { "LDAP_OPT_X_TLS_PROTOCOL_TLS1_0", LDAP_OPT_X_TLS_PROTOCOL_TLS1_0 },
  { "LDAP_OPT_X_TLS_PROTOCOL_TLS1_1", LDAP_OPT_X_TLS_PROTOCOL_TLS1_1 },
  { "LDAP_OPT_X_TLS_PROTOCOL_TLS1_2", LDAP_OPT_X_TLS_PROTOCOL_TLS1_2 },
  { NULL,                             -1 }
};

/* XXX Union initialization may not be supported by older compilers... */
typedef union Ldap_option_value {
  char *strval;			/* If type is LDAP_OPT_STR. */
  int intval;			/* LDAP_OPT_INT */
  void *boolval;		/* LDAP_OPT_BOOL */
  char *enumval;		/* LDAP_OPT_ENUM */
} Ldap_option_value;

/*
 * These are OpenLDAP options that are (or may be) needed to be changed
 * for DACS.  Options not in this table are unknown to DACS but may be added.
 *
 * Any option that is not initialized in this table or set by a DACS
 * configuration OPTION directive causes OpenLDAP to follow its default
 * behaviour, which can depend on environment variables (e.g., set
 * through Apache SetEnv directives), or OpenLDAP configuration or its defaults.
 *
 * OpenLDAP options are applied only for the one connection (or
 * two connections, in the case of indirect binds) established for
 * one instance of local_ldap_authenticate; they are in no way global or
 * sticky.
 *
 * An "enumerated value" (e.g., LDAP_OPT_X_TLS_NEVER) is mapped to the
 * corresponding OpenLDAP integer value; the integer value to use is kept in
 * value.intval, the default symbolic value is in default_value.enumval,
 * and the user configured symbol value is in config_str.
 *
 * XXX this is somewhat fragile since it depends on OpenLDAP naming and
 * definitions, which could change.
 */
typedef struct Ldap_option_tab {
  int opt;							/* OpenLDAP's option symbol's value. */
  char *name;						/* OpenLDAP option name with prefix. */
  int always_set;					/* If zero, set only if was_overridden. */
  int was_overridden;				/* Did DACS config set this value? */
  Ldap_option_type type;			/* Option value type. */
  char *enum_val_str;				/* If an enum, its value as a string. */
  Ldap_option_value default_value;	/* DACS default value. */
  Ldap_option_value value;			/* Value requested by DACS config. */
} Ldap_option_tab;

/*
 * Only options currently of interest to us, not all possible options.
 * Extend as necessary.
 * LDAP_OPT_NETWORK_TIMEOUT, LDAP_OPT_DEREF, LDAP_OPT_SIZELIMIT,
 * LDAP_OPT_URI, LDAP_OPT_X_TLS_CRLCHECK, LDAP_OPT_X_TLS_CRLFILE
 */
static Ldap_option_tab ldap_option_tab[17] = {
  /* TLS specific options. */
  /* Takes same values as X_TLS_REQUIRE_CERT but this is ldo_tls_mode. */
  { LDAP_OPT_X_TLS, "LDAP_OPT_X_TLS",
	1, 0, LDAP_OPT_ENUM, NULL, { .enumval = "LDAP_OPT_X_TLS_HARD" }, { NULL } },
  { LDAP_OPT_X_TLS_CACERTDIR, "LDAP_OPT_X_TLS_CACERTDIR",
	0, 0, LDAP_OPT_STR, NULL, { .strval = NULL }, { NULL } },
  { LDAP_OPT_X_TLS_CACERTFILE, "LDAP_OPT_X_TLS_CACERTFILE",
	0, 0, LDAP_OPT_STR, NULL, { .strval = NULL }, { NULL } },
  { LDAP_OPT_X_TLS_CERTFILE, "LDAP_OPT_X_TLS_CERTFILE",
	0, 0, LDAP_OPT_STR, NULL, { .strval = NULL }, { NULL } },
  { LDAP_OPT_X_TLS_CIPHER_SUITE, "LDAP_OPT_X_TLS_CIPHER_SUITE",
	0, 0, LDAP_OPT_STR, NULL, { .strval = NULL }, { NULL } },
  { LDAP_OPT_X_TLS_KEYFILE, "LDAP_OPT_X_TLS_KEYFILE",
	0, 0, LDAP_OPT_STR, NULL, { .strval = NULL }, { NULL } },
  { LDAP_OPT_X_TLS_PROTOCOL_MIN, "LDAP_OPT_X_TLS_PROTOCOL_MIN",
	0, 0, LDAP_OPT_INT, NULL, { .intval = LDAP_OPT_X_TLS_PROTOCOL_TLS1_2 },
	{ NULL } },
  { LDAP_OPT_X_TLS_RANDOM_FILE, "LDAP_OPT_X_TLS_RANDOM_FILE",
	0, 0, LDAP_OPT_STR, NULL, { .strval = NULL }, { NULL } },
  /* Takes same values as X_TLS but this is ldo_tls_require_cert. */
  { LDAP_OPT_X_TLS_REQUIRE_CERT, "LDAP_OPT_X_TLS_REQUIRE_CERT",
	1, 0, LDAP_OPT_ENUM, NULL, { .enumval = "LDAP_OPT_X_TLS_NEVER" }, { NULL } },

  /* General LDAP options. */
  { LDAP_OPT_DEBUG_LEVEL, "LDAP_OPT_DEBUG_LEVEL",
	0, 0, LDAP_OPT_INT, NULL, { .intval = 0 }, { NULL } },
  { LDAP_OPT_PROTOCOL_VERSION, "LDAP_OPT_PROTOCOL_VERSION",
	1, 0, LDAP_OPT_INT, NULL, { .intval = LDAP_VERSION3 }, { NULL } },
  { LDAP_OPT_REFERRALS, "LDAP_OPT_REFERRALS",
	1, 0, LDAP_OPT_BOOL, NULL, { .boolval = LDAP_OPT_OFF }, { NULL } },
  { LDAP_OPT_RESTART, "LDAP_OPT_RESTART",
	0, 0, LDAP_OPT_BOOL, NULL, { .boolval = LDAP_OPT_OFF }, { NULL } },
  /* A non-negative number of seconds for searches; zero means unlimited. */
  { LDAP_OPT_TIMELIMIT, "LDAP_OPT_TIMELIMIT",
	0, 0, LDAP_OPT_INT, NULL, { .intval = 0 }, { NULL } },
  /* A non-negative number of seconds for responses; -1 means unlimited. */
  { LDAP_OPT_TIMEOUT, "LDAP_OPT_TIMEOUT",
	0, 0, LDAP_OPT_INT, NULL, { .intval = -1 }, { NULL } },

  { -1, NULL,
	0, 0, LDAP_OPT_NONE, NULL, { .intval = 0 }, { NULL } }
};

/* Map a symbolic (enum) name into its integer value. */
static Ldap_option_enum *
lookup_ldap_option_enum(char *name)
{
 Ldap_option_enum *et;

  for (et = ldap_option_enums; et->name != NULL; et++) {
	if (strcaseeq(et->name, name))
	  return(et);
  }

  return(NULL);
}

/*
 * Set the default OpenLDAP debug level, which may still be overridden
 * by a DACS OPTION, so call this when initializing.
 */
static int
config_ldap_debug_option(void)
{
  Ldap_option_tab *t;

  if (trace_level <= 1)
	return(0);

  for (t = ldap_option_tab; t->name != NULL; t++) {
	if (strcaseeq(t->name, "LDAP_OPT_DEBUG_LEVEL")) {
	  t->default_value.intval = 1;
	  return(0);
	}
  }

  return(-1);
}

/*
 * Initialize ldap_option_tab so that default values are used.
 * This may be called to reinitialize the table, which may be useful
 * when testing;
 * Return 0 if ok, -1 if an error occurs.
 */
static int
init_ldap_options(void)
{
  Ldap_option_tab *t;

  for (t = ldap_option_tab; t->name != NULL; t++) {
	t->was_overridden = 0;
	t->enum_val_str = NULL;

	switch (t->type) {
	case LDAP_OPT_STR:
	  t->value.strval = NULL;
	  break;

	case LDAP_OPT_INT:
	  t->value.intval = t->default_value.intval;
	  break;

	case LDAP_OPT_BOOL:
	  t->value.boolval = t->default_value.boolval;
	  break;

	case LDAP_OPT_ENUM:
	  {
		Ldap_option_enum *e;

		if ((e = lookup_ldap_option_enum(t->default_value.enumval)) == NULL)
		  return(-1);
		t->value.intval = e->intval;
		t->enum_val_str = strdup(t->default_value.enumval);

		break;
	  }

	case LDAP_OPT_NONE:
	  break;
	}
  }

  config_ldap_debug_option();

  return(0);
}

static int
dump_ldap_options(void)
{
  int i, rc;
  Ds ds;
  Ldap_option_tab *t;

  ds_init(&ds);
  for (t = ldap_option_tab; t->name != NULL; t++) {
	char *action;

	ds_reinit(&ds);
	ds_asprintf(&ds, "(%s, will %s)",
				t->was_overridden ? "config" : "default",
				(t->always_set || t->was_overridden) ? "set" : "ignore");
	action = ds_buf(&ds);

	switch (t->type) {
	case LDAP_OPT_STR:
	  if (t->value.strval == NULL)
		log_msg((LOG_DEBUG_LEVEL, "%s is NULL %s", t->name, action));
	  else
		log_msg((LOG_DEBUG_LEVEL, "%s is \"%s\" %s",
				 t->name, t->value.strval, action));
	  break;

	case LDAP_OPT_INT:
	  log_msg((LOG_DEBUG_LEVEL, "%s is %d %s",
			   t->name, t->value.intval, action));
	  break;

	case LDAP_OPT_BOOL:
	  log_msg((LOG_DEBUG_LEVEL, "%s is %s %s",
			   t->name,
			   t->value.boolval ? "LDAP_OPT_ON" : "LDAP_OPT_OFF", action));
	  break;

	case LDAP_OPT_ENUM:
	  log_msg((LOG_DEBUG_LEVEL, "%s is %s (== %d) %s",
			   t->name, t->enum_val_str, t->value.intval, action));
	  break;

	default:
	  log_msg((LOG_ERROR_LEVEL, "Internal error: invalid option for '%s'",
			   t->name));
	  return(-1);
	}
  }

  if (log_would_log(LOG_DEBUG_LEVEL)) {
	Ds ds_enumlist, ds_varlist;
	Ldap_option_enum *et;

	ds_init(&ds_varlist);
	for (t = ldap_option_tab; t->name != NULL; t++) {
	  ds_asprintf(&ds_varlist, "%s%s", (ds_len(&ds_varlist) == 0) ? "" : ", ",
				  t->name);
	}
	log_msg((LOG_DEBUG_LEVEL, "Defined LDAP OPTION variables: %s",
			 ds_buf(&ds_varlist)));

	ds_init(&ds_enumlist);
	for (et = ldap_option_enums; et->name != NULL; et++) {
	  ds_asprintf(&ds_enumlist, "%s%s",
				  (ds_len(&ds_enumlist) == 0) ? "" : ", ", et->name);
	}

	log_msg((LOG_DEBUG_LEVEL, "Defined LDAP OPTION enums: %s",
			 ds_buf(&ds_enumlist)));
  }

  return(0);
}

/*
 * Configure an OpenLDAP option to override all defaults.
 * If a particular option is attempted to be set more than once, only the
 * first has any effect (the others are ignored).
 *
 * Return 1 if option NAME=VALUE was set, 0 if it can be safely ignored
 * (e.g., it is a duplicate), or -1 if an error occurred.
 */
static int
config_ldap_option(char *name, char *value)
{
  Ldap_option_tab *t;
  Ldap_option_enum *et;

  if (name == NULL || *name == '\0')
	return(-1);

  for (t = ldap_option_tab; t->name != NULL; t++) {
	if (strcaseeq(t->name, name)) {
	  if (t->was_overridden) {
		log_msg((LOG_WARN_LEVEL,
				 "Ignoring duplicate LDAP option: '%s'", name));
		return(0);
	  }

	  switch (t->type) {
	  case LDAP_OPT_STR:
		if (value == NULL) {
		  t->value.strval = NULL;
		  log_msg((LOG_DEBUG_LEVEL, "Set string option %s to NULL", name));
		}
		else {
		  t->value.strval = strdup(value);
		  log_msg((LOG_DEBUG_LEVEL, "Set string option %s to \"%s\"",
				   name, value));
		}
		break;

	  case LDAP_OPT_INT:
		if (value == NULL)
		  t->value.intval = 0;
		else {
		  if (strnum(value, STRNUM_I, &t->value.intval) == -1) {
			log_msg((LOG_ERROR_LEVEL,
					 "Invalid integer option value for '%s'", name));
			return(-1);
		  }
		}
		log_msg((LOG_DEBUG_LEVEL,
				 "Set integer option %s to %d", name, t->value.intval));
		break;

	  case LDAP_OPT_BOOL:
		if (value == NULL || strcaseeq(value, "on"))
		  t->value.boolval = LDAP_OPT_ON;
		else
		  t->value.boolval = LDAP_OPT_OFF;
		log_msg((LOG_DEBUG_LEVEL,
				 "Set boolean option %s to \"%s\"",
				 name, (t->value.boolval) ? "on" : "off"));
		break;

	  case LDAP_OPT_ENUM:
		if (value == NULL) {
		  log_msg((LOG_ERROR_LEVEL, "Invalid enum option %s: no value", name));
		  return(-1);
		}

		if ((et = lookup_ldap_option_enum(value)) == NULL) {
		  log_msg((LOG_ERROR_LEVEL, "Unrecognized enum option %s: \"%s\"",
				   name, value));
		  return(-1);
		}

		t->enum_val_str = strdup(value);
		t->value.intval = et->intval;
		log_msg((LOG_DEBUG_LEVEL,
				 "Set enum option %s to %s (== %d)",
				 name, t->enum_val_str, t->value.intval));
		break;

	  default:
		log_msg((LOG_ERROR_LEVEL, "Internal error: invalid option for '%s'",
				 name));
		return(-1);
	  }

	  t->was_overridden = 1;
	  break;
	}
  }

  if (t->name == NULL) {
	log_msg((LOG_ERROR_LEVEL, "Unrecognized LDAP option: '%s'", name));
	return(-1);
  }

  return(1);
}

/*
 * Set all OpenLDAP options that are either DACS defaults or from DACS config.
 * init_ldap_options() and (if any options have been configured through DACS)
 * config_ldap_option() should have already been called to initialize and
 * change option values.
 *
 * It is assumed that an option set via OpenLDAP's ldap_set_option(3) overrides
 * the corresponding environment variable and also any settings in configuration
 * files that might be read by OpenLDAP (see ldap.conf(5)).
 */
static int
set_ldap_options(LDAP *ld)
{
  int int_option, rc;
  char *str_option;
  void *bool_option;
  Ldap_option_tab *t;

  for (t = ldap_option_tab; t->name != NULL; t++) {
	if (!t->was_overridden && !t->always_set)
	  continue;

	switch (t->type) {
	case LDAP_OPT_STR:
	  rc = ldap_set_option(ld, t->opt, t->value.strval);
	  break;

	case LDAP_OPT_INT:
	  rc = ldap_set_option(ld, t->opt, &t->value.intval);
	  break;

	case LDAP_OPT_BOOL:
	  rc = ldap_set_option(ld, t->opt, t->value.boolval);
	  break;

	case LDAP_OPT_ENUM:
	  rc = ldap_set_option(ld, t->opt, &t->value.intval);
	  break;

	default:
	  log_msg((LOG_ERROR_LEVEL, "Internal error: invalid option for '%s'",
			   t->name));
	  return(-1);
	}

	if (rc != LDAP_OPT_SUCCESS) {
	  log_msg((LOG_ERROR_LEVEL,
			   "Cannot set option %s: %s", t->name, ldap_err2string(rc)));
	  return(-1);
	}
  }

  return(0);
}

/*
 * Initialize and open a connection to an LDAP server according to URL.
 * Apply DACS default and DACS configured options.
 */
static LDAP *
do_init(Uri *url)
{
  int debug, ldap_option, rc;
  LDAP *ld;

#ifdef OLD
  int port;
  char *host;

  host = url->host;
  port = url->port;

  if ((ld = ldap_init(host, port)) == NULL) {
	log_msg((LOG_ERROR_LEVEL, "ldap_init to %s:%d failed", host, port));
	return(NULL);
  }
#else
  char *sa;

  sa = uri_scheme_authority(url);
  ld = NULL;
  if ((rc = ldap_initialize(&ld, sa)) != LDAP_SUCCESS) {
	log_msg((LOG_ERROR_LEVEL, "ldap_initialize to \"%s\" failed: %s",
			 sa, ldap_err2string(rc)));
	return(NULL);
  }
  log_msg((LOG_DEBUG_LEVEL, "ldap_initialize: \"%s\"", sa));

#ifdef NOTDEF
  if (log_would_log(LOG_DEBUG_LEVEL)) {
	char *opt_cacertfile = NULL;
	char *env_tls_cacert = NULL;
	char *env_ldaptls_cacert = NULL;

	rc = ldap_get_option(ld, LDAP_OPT_X_TLS_CACERTFILE, &opt_cacertfile);
	if (rc != LDAP_OPT_SUCCESS) {
	  log_msg((LOG_ERROR_LEVEL,
			   "Cannot get LDAP_OPT_X_TLS_CACERTFILE: %s",
			   ldap_err2string(rc)));
	  return(NULL);
	}
	if (opt_cacertfile)
	  log_msg((LOG_DEBUG_LEVEL, "option LDAP_OPT_X_TLS_CACERTFILE is \"%s\"",
			   opt_cacertfile));
	else
	  log_msg((LOG_DEBUG_LEVEL, "option LDAP_OPT_X_TLS_CACERTFILE is unset"));

	/* This should not be present. */
	env_tls_cacert = getenv("TLS_CACERT");
	if (env_tls_cacert)
	  log_msg((LOG_DEBUG_LEVEL, "envar TLS_CACERT is \"%s\"", env_tls_cacert));
	else
	  log_msg((LOG_DEBUG_LEVEL, "envar TLS_CACERT is unset"));

	/* This may be present - OpenLDAP expects the prefix "LDAP". */
	env_ldaptls_cacert = getenv("LDAPTLS_CACERT");
	if (env_ldaptls_cacert)
	  log_msg((LOG_DEBUG_LEVEL,
			   "envar LDAPTLS_CACERT is \"%s\"", env_ldaptls_cacert));
	else
	  log_msg((LOG_DEBUG_LEVEL, "envar LDAPTLS_CACERT is unset"));
  }
#endif
#endif

  /*
   * Apply the options we need both globally (for the duration of this process)
   * and for this connection to the LDAP server.
   */
  if (set_ldap_options(NULL) == -1)
	return(NULL);
  if (set_ldap_options(ld) == -1)
	return(NULL);

  dump_ldap_options();

  return(ld);
}

/*
 * Bind to URL using PASSWORD.
 * Return 1 if successful, 0 if unsuccessful but another bind method can
 * be attempted, -1 on error.
 */
int
local_ldap_auth_by_dn(Uri *url, char *username, char *password,
					  struct timeval *timeout, Dsvec *avas)
{
  int attrsonly, n, scope;
  char *user_dn;
  char *attrs[100];
  LDAP *ld;

  if ((ld = do_init(url)) == NULL)
	return(-1);

  if (*url->path == '/')
	user_dn = url->path + 1;
  else
	user_dn = url->path;

  log_msg((LOG_TRACE_LEVEL, "Authenticating by binding to \"%s\"", user_dn));

  if (do_bind(ld, user_dn, password) == -1) {
	log_msg((LOG_ERROR_LEVEL, "Bind to \"%s\" failed", user_dn));
	ldap_unbind(ld);
	return(0);
  }
  log_msg((LOG_DEBUG_LEVEL, "Bind to \"%s\" succeeded", user_dn));

  if (avas != NULL) {
	scope = LDAP_SCOPE_BASE;
	attrsonly = 0;
	attrs[0] = "*";		/* Get all attributes */
	attrs[1] = NULL;
	n = fetch_entry(ld, user_dn, scope, NULL, NULL,
					attrs, attrsonly, NULL, NULL, timeout, -1, NULL, avas);
	if (n == -1) {
	  log_msg((LOG_ERROR_LEVEL, "Cannot read entry for DN %s", user_dn));
	  ldap_unbind(ld);
	  return(-1);
	}
	else if (n > 1) {
	  log_msg((LOG_WARN_LEVEL, "Too many results for DN %s", user_dn));
	  ldap_unbind(ld);
	  return(-1);
	}
	log_msg((LOG_DEBUG_LEVEL, "Read entry for DN %s", user_dn));
  }

  ldap_unbind(ld);

  return(1);
}

/*
 * Bind as ADMIN_URL using ADMIN_PASSWORD (if provided) and search
 * for a (unique) entry rooted at ROOT_DN, applying FILTER and TIMEOUT to
 * the search.
 * If the search is successful, attempt to bind to that entry using PASSWORD
 * and return that entry's AVAs.
 * Return 1 if successful, 0 if unsuccessful but another bind method can
 * be attempted, -1 on error.
 */
int
local_ldap_auth_by_attr(Uri *admin_url, char *admin_password,
						char *root_dn, char *filter, char *password,
						struct timeval *timeout, Dsvec *avas)
{
  int attrsonly, n, scope;
  char *admin_bind_dn, *user_dn;
  char *attrs[1];
  LDAP *ld;

  if ((ld = do_init(admin_url)) == NULL)
	return(-1);

  if (*admin_url->path == '/')
	admin_bind_dn = admin_url->path + 1;
  else
	admin_bind_dn = admin_url->path;

  if (do_bind(ld, admin_bind_dn, admin_password) == -1) {
	log_msg((LOG_ERROR_LEVEL, "Admin bind to %s failed", admin_bind_dn));
	ldap_unbind(ld);
	return(-1);
  }

  log_msg((LOG_DEBUG_LEVEL, "Bind to \"%s\" succeeded", admin_bind_dn));

  scope = LDAP_SCOPE_SUBTREE;
  attrsonly = 0;
  attrs[0] = NULL;

  log_msg((LOG_TRACE_LEVEL,
		   "Authenticating by searching under \"%s\" for \"%s\"",
		   root_dn, filter));

  n = fetch_entry(ld, root_dn, scope, NULL, filter,
				  attrs, attrsonly, NULL, NULL, timeout, -1, &user_dn, avas);
  if (n < 1) {
	log_msg((LOG_ERROR_LEVEL, "Search at %s for %s failed", root_dn, filter));
	ldap_unbind(ld);
	return(0);
  }
  else if (n > 1) {
	log_msg((LOG_ERROR_LEVEL, "Too many search results at %s for %s",
			 root_dn, filter));
	ldap_unbind(ld);
	return(-1);
  }

  log_msg((LOG_TRACE_LEVEL, "Search found DN \"%s\"", user_dn));

  ldap_unbind(ld);

  if ((ld = do_init(admin_url)) == NULL)
	return(-1);

  log_msg((LOG_DEBUG_LEVEL, "Authenticating by binding to DN: %s", user_dn));

  if (!local_ldap_roles) {
	if (do_bind(ld, user_dn, password) == -1) {
	  log_msg((LOG_ERROR_LEVEL, "User bind to %s failed", user_dn));
	  ldap_unbind(ld);
	  return(-1);
	}
  }

  ldap_unbind(ld);

  return(1);
}

static char *
make_roles(Kwv_pair *roles_selectors, Kwv *attrs)
{
  int n;
  Ds ds;
  Kwv_pair *k, *rs;

  ds_init(&ds);

  /*
   * Evaluate each LDAP_ROLES_SELECTOR* directive, iterating through each
   * attribute of the entry (ATTRS) and passing the variable ${LDAP::attrname}
   * to the name of the attribute and ${LDAP::attrvalue} to its value.
   */
  n = 0;
  for (rs = roles_selectors; rs != NULL; rs = rs->next) {
	Kwv_iter *iter;

	log_msg((LOG_TRACE_LEVEL, "Trying roles selector: %s", rs->val));
	if ((iter = kwv_iter_begin(attrs, NULL)) != NULL) {
	  for (k = kwv_iter_first(iter); k != NULL; k = kwv_iter_next(iter)) {
		char *p;
		Acs_expr_ns_arg ns_args[3];

		if (streq(k->name, "attrname") || streq(k->name, "attrvalue"))
		  continue;
		if (k->val == NULL)
		  continue;

		kwv_replace(attrs, "attrname", k->name);
		kwv_replace(attrs, "attrvalue", k->val);
		log_msg((LOG_TRACE_LEVEL, "Trying attr: %s=\"%s\"", k->name, k->val));
		ns_args[0].name = "LDAP";
		ns_args[0].kwv = attrs;
		ns_args[1].name = dacs_conf->conf_var_ns->ns;
		ns_args[1].kwv = dacs_conf->conf_var_ns->kwv;
		ns_args[2].name = NULL;
		ns_args[2].kwv = NULL;
		acs_expr_string(rs->val, ns_args, &p);
		if (p == NULL || *p == '\0')
		  continue;
		if (is_valid_role_str(p)) {
		  ds_asprintf(&ds, "%s%s", n ? "," : "", p);
		  n++;
		  log_msg((LOG_TRACE_LEVEL, "Selected role attr value: %s", p));
		}
	  }
	  kwv_iter_end(iter);
	  kwv_delete(attrs, "attrname");
	  kwv_delete(attrs, "attrvalue");
	}
  }

  if (n == 0)
	return(NULL);

  return(ds_buf(&ds));
}

int
main(int argc, char **argv)
{
  int debug_flag, direct_bind, indirect_bind, emitted_dtd, i, st, success;
  char *bind_method, *errmsg, *jurisdiction, *p, *username, *password, *aux;
  char *mapped_username, *role_str, *timeout_secs;
  char *auth_ldap_username_url;
  char *auth_ldap_username_expr;
  char *auth_ldap_admin_url;
  char *auth_ldap_admin_password;
  char *auth_ldap_search_root_dn, *auth_ldap_search_filter;
  struct timeval timeval, *tv;
  Auth_reply_ok ok;
  DACS_app_type app_type;
  Dsvec *avas;
  Kwv *kwv, *kwv_attrs;
  Kwv_pair *v;
  Uri *url;

  if (argv[1] != NULL && streq(argv[1], "-options")) {
	dump_ldap_options();
	exit(0);
  }

  debug_flag = 0;
  errmsg = "Internal error";
  emitted_dtd = 0;
  username = password = aux = jurisdiction = NULL;

  /*
   * Depending on what name is used to invoke the program, function as
   * either local_ldap_authenticate or local_ldap_roles.
   * This seems sensible because they share almost all of the code.
   */
  if ((p = current_uri_prog()) == NULL
	  || streq(p, "local_ldap_authenticate")) {
	log_module_name = "local_ldap_auth";
	local_ldap_roles = 0;
  }
  else if (streq(p, "local_ldap_roles")) {
	log_module_name = "local_ldap_roles";
	local_ldap_roles = 1;
  }
  else {
	errmsg = "An unrecognized URI was used for this service";
	goto fail;
  }

  if (debug_flag)
	app_type = DACS_STANDALONE_NOARGS;
  else
	app_type = DACS_LOCAL_SERVICE;

  if (dacs_init(app_type, &argc, &argv, &kwv, &errmsg) == -1) {
	/* If we fail here, we may not have a DTD with which to reply... */
  fail:
	if (password != NULL)
	  strzap(password);
	if (aux != NULL)
	  strzap(aux);

	if (local_ldap_roles)
	  emit_roles_reply_failed(stdout, errmsg);
	else {
	  if (emitted_dtd) {
		printf("%s\n", make_xml_auth_reply_failed(NULL, NULL));
		emit_xml_trailer(stdout);
	  }
	}

	if (errmsg != NULL)
	  log_msg((LOG_ERROR_LEVEL, "Failed: reason=%s", errmsg));

	exit(1);
  }

  if (debug_flag) {
	Log_desc *ld;

	ld = log_init(NULL, 0, NULL, "local_ldap_auth", LOG_NONE_LEVEL, NULL);
	log_set_level(ld, LOG_DEBUG_LEVEL);
	log_set_desc(ld, LOG_ENABLED);
  }

  if (init_ldap_options() == -1) {
	errmsg = "Internal error while initializing LDAP options";
	goto fail;
  }

  if (!local_ldap_roles) {
	/* This must go after initialization. */
	emitted_dtd = emit_xml_header(stdout, "auth_reply");
  }

  if (argc > 1) {
	errmsg = "Usage: unrecognized parameter";
	goto fail;
  }

  /*
   * If called as a roles module, the username should be passed as
   * DACS_USERNAME (an authenticated name) but if called as an authentication
   * module, the username should be passed as USERNAME (an unauthenticated
   * name).  It is an error if both parameters are present.
   */
  for (i = 0; i < kwv->nused; i++) {
	if (!local_ldap_roles
		&& streq(kwv->pairs[i]->name, "USERNAME") && username == NULL)
	  username = kwv->pairs[i]->val;
	else if (local_ldap_roles
			 && streq(kwv->pairs[i]->name, "DACS_USERNAME") && username == NULL)
	  username = kwv->pairs[i]->val;
	else if (streq(kwv->pairs[i]->name, "PASSWORD") && password == NULL)
	  password = kwv->pairs[i]->val;
	else if (streq(kwv->pairs[i]->name, "AUXILIARY") && aux == NULL)
	  aux = kwv->pairs[i]->val;
	else if (streq(kwv->pairs[i]->name, "DACS_JURISDICTION")
			 && jurisdiction == NULL)
	  jurisdiction = kwv->pairs[i]->val;
	else if (streq(kwv->pairs[i]->name, "DACS_VERSION"))
	  ;
	else if (strprefix(kwv->pairs[i]->name, LDAP_OPTION_PREFIX) != NULL) {
	  if (config_ldap_option(kwv->pairs[i]->name, kwv->pairs[i]->val) == -1) {
		errmsg = "Invalid LDAP option";
		goto fail;
	  }
	}
	else
	  log_msg((LOG_TRACE_LEVEL, "Parameter: '%s'", kwv->pairs[i]->name));
  }

  /* Verify that we're truly responsible for DACS_JURISDICTION. */
  if (dacs_verify_jurisdiction(jurisdiction) == -1) {
	errmsg = "Missing or incorrect DACS_JURISDICTION";
	goto fail;
  }

  direct_bind = 0;
  indirect_bind = 0;
  if ((bind_method = kwv_lookup_value(kwv, "LDAP_BIND_METHOD")) == NULL) {
	errmsg = "No LDAP_BIND_METHOD";
	goto fail;
  }
  if (strcaseeq(bind_method, "direct"))
	direct_bind = 1;
  else if (strcaseeq(bind_method, "indirect"))
	indirect_bind = 1;
  else if (strcaseeq(bind_method, "both"))
	direct_bind = indirect_bind = 1;
  else {
	errmsg = ds_xprintf("Unrecognized LDAP_BIND_METHOD type: \"%s\"",
						bind_method);;
	goto fail;
  }

  avas = dsvec_init(NULL, sizeof(Ava));

  if ((timeout_secs = kwv_lookup_value(kwv, "LDAP_TIMEOUT_SECS")) != NULL) {
	unsigned long ul;

	if (strnum(timeout_secs, STRNUM_UL, &ul) == -1) {
	  ul = 0;
	  log_msg((LOG_WARN_LEVEL,
			   "Invalid LDAP_TIMEOUT_SECS, ignored: \"%s\"", timeout_secs));
	}
	timeval.tv_sec = ul;
	timeval.tv_usec = 0;
	tv = &timeval;
  }
  else
	tv = NULL;

  success = 0;
  if (direct_bind) {
	char *p_e;
	Acs_expr_ns_arg ns_args[3];

	log_msg((LOG_TRACE_LEVEL, "Trying direct bind method"));
	p_e = kwv_lookup_value(kwv, "LDAP_USERNAME_URL*");
	p = kwv_lookup_value(kwv, "LDAP_USERNAME_URL");
	if (p == NULL && p_e == NULL) {
	  errmsg = "No LDAP_USERNAME_URL or LDAP_USERNAME_URL* available";
	  goto fail;
	}
	if (p != NULL && p_e != NULL) {
	  errmsg = "Either LDAP_USERNAME_URL or LDAP_USERNAME_URL* is required";
	  goto fail;
	}

	if (p_e != NULL) {
	  ns_args[0].name = "Args";
	  ns_args[0].kwv = kwv;
	  ns_args[1].name = dacs_conf->conf_var_ns->ns;
	  ns_args[1].kwv = dacs_conf->conf_var_ns->kwv;
	  ns_args[2].name = NULL;
	  ns_args[2].kwv = NULL;
	  acs_expr_string(p_e, ns_args, &auth_ldap_username_url);
	}
	else
	  auth_ldap_username_url = p;

	if ((url = uri_parse(auth_ldap_username_url)) == NULL) {
	  errmsg = ds_xprintf("Cannot parse LDAP_USERNAME_URL: \"%s\"",
						  auth_ldap_username_url);
	  goto fail;
	}
	if (url->scheme != NULL && strcaseeq(url->scheme, "ldap")) {
	  if (url->port_given == NULL)
		url->port = LDAP_DEFAULT_PORT;
	}
	else if (url->scheme != NULL && strcaseeq(url->scheme, "ldaps")) {
	  if (url->port_given == NULL)
		url->port = LDAP_DEFAULT_SSL_PORT;
	}
	else {
	  errmsg = "Unrecognized scheme in LDAP_USERNAME_URL";
	  goto fail;
	}

#ifdef OLD
	if (strcaseeq(url->scheme, "ldaps")) {
	  errmsg = "ldaps scheme is unsupported in LDAP_USERNAME_URL";
	  goto fail;
	}
#endif

	log_msg((LOG_TRACE_LEVEL, "Direct bind: %s", url->uri));
	st = local_ldap_auth_by_dn(url, username, password, tv, avas);
	if (st == -1) {
	  errmsg = "Username/Password/Aux incorrect";
	  goto fail;
	}
	if (st == 1) {
	  log_msg((LOG_DEBUG_LEVEL, "Authentication was successful"));
	  success = 1;
    }
  }

  if (!success && indirect_bind) {
	char *p_e;
	Acs_expr_ns_arg ns_args[3];

	log_msg((LOG_TRACE_LEVEL, "Trying indirect bind method"));
	if ((p = kwv_lookup_value(kwv, "LDAP_ADMIN_URL")) == NULL) {
	  errmsg = "No LDAP_ADMIN_URL available";
	  goto fail;
	}
	auth_ldap_admin_url = p;

	if ((p = kwv_lookup_value(kwv, "LDAP_ADMIN_PASSWORD")) == NULL) {
	  p = "";
	  log_msg((LOG_DEBUG_LEVEL, "No LDAP_ADMIN_PASSWORD provided"));
	}
	auth_ldap_admin_password = p;

	if ((p = kwv_lookup_value(kwv, "LDAP_SEARCH_ROOT_DN")) == NULL) {
	  p = "";
	  log_msg((LOG_DEBUG_LEVEL, "No LDAP_SEARCH_ROOT_DN provided"));
	}
	auth_ldap_search_root_dn = p;

	p_e = kwv_lookup_value(kwv, "LDAP_SEARCH_FILTER*");
	p = kwv_lookup_value(kwv, "LDAP_SEARCH_FILTER");
	if (p == NULL && p_e == NULL) {
	  errmsg = "No LDAP_SEARCH_FILTER or LDAP_SEARCH_FILTER* available";
	  goto fail;
	}
	if (p != NULL && p_e != NULL) {
	  errmsg = "Either LDAP_SEARCH_FILTER or LDAP_SEARCH_FILTER* is required";
	  goto fail;
	}

	if (p_e != NULL) {
	  ns_args[0].name = "Args";
	  ns_args[0].kwv = kwv;
	  ns_args[1].name = dacs_conf->conf_var_ns->ns;
	  ns_args[1].kwv = dacs_conf->conf_var_ns->kwv;
	  ns_args[2].name = NULL;
	  ns_args[2].kwv = NULL;
	  acs_expr_string(p_e, ns_args, &auth_ldap_search_filter);
	  if (auth_ldap_search_filter == NULL) {
		errmsg = "Cannot form search filter";
		goto fail;
	  }
	}
	else
	  auth_ldap_search_filter = p;

	if ((url = uri_parse(auth_ldap_admin_url)) == NULL) {
	  errmsg = ds_xprintf("Cannot parse LDAP_ADMIN_URL: \"%s\"",
						  auth_ldap_admin_url);
	  goto fail;
	}
	if (url->scheme != NULL && strcaseeq(url->scheme, "ldap")) {
	  if (url->port_given == NULL)
		url->port = LDAP_DEFAULT_PORT;
	}
	else if (url->scheme != NULL && strcaseeq(url->scheme, "ldaps")) {
	  if (url->port_given == NULL)
		url->port = LDAP_DEFAULT_SSL_PORT;
	}
	else {
	  errmsg = "Unrecognized scheme in LDAP_ADMIN_URL";
	  goto fail;
	}
	log_msg((LOG_DEBUG_LEVEL,
			 "auth_ldap_admin_url=\"%s\", auth_ldap_search_root_dn=\"%s\"",
			 auth_ldap_admin_url, auth_ldap_search_root_dn));

#ifdef OLD
	if (strcaseeq(url->scheme, "ldaps")) {
	  errmsg = "ldaps scheme is unsupported in LDAP_ADMIN_URL";
	  goto fail;
	}
#endif

	log_msg((LOG_TRACE_LEVEL, "Indirect bind: %s", url->uri));
	st = local_ldap_auth_by_attr(url, auth_ldap_admin_password,
								 auth_ldap_search_root_dn,
								 auth_ldap_search_filter, password, tv, avas);
 	if (st == -1) {
	  errmsg = "Username/Password/Aux incorrect";
	  goto fail;
	}
	if (st == 1) {
	  log_msg((LOG_DEBUG_LEVEL, "Authentication was successful"));
	  success = 1;
    }
  }

  if (!success) {
	if (username != NULL)
	  errmsg = ds_xprintf("Cannot bind to directory as user \"%s\"",
						  username);
	else
	  errmsg = "Cannot bind to directory";
	goto fail;
  }

  if (password != NULL)
	strzap(password);
  if (aux != NULL)
	strzap(aux);

  if (!local_ldap_roles) {
	ok.username = NULL;
	/* If this wasn't specified, dacs_authenticate will use the default. */
	ok.lifetime = kwv_lookup_value(kwv, "CREDENTIALS_LIFETIME_SECS");
	ok.roles_reply = NULL;
  }

  kwv_attrs = kwv_init(40);
  kwv_attrs->dup_mode = KWV_ALLOW_DUPS;
  for (i = 0; i < dsvec_len(avas); i++) {
	Ava *a;

	a = dsvec_ptr(avas, i, Ava *);
	log_msg((LOG_TRACE_LEVEL, "attr=\"%s\" value=\"%s\"",
			 a->attr_name, a->attr_value));
	kwv_add(kwv_attrs, a->attr_name, a->attr_value);
  }

  if ((p = kwv_lookup_value(kwv, "LDAP_USERNAME_EXPR*")) == NULL) {
	log_msg((LOG_DEBUG_LEVEL, "No LDAP_USERNAME_EXPR* directive"));
	mapped_username = username;
  }
  else {
	Acs_expr_ns_arg ns_args[3];

	auth_ldap_username_expr = p;
	ns_args[0].name = "LDAP";
	ns_args[0].kwv = kwv_attrs;
	ns_args[1].name = dacs_conf->conf_var_ns->ns;
	ns_args[1].kwv = dacs_conf->conf_var_ns->kwv;
	ns_args[2].name = NULL;
	ns_args[2].kwv = NULL;
	acs_expr_string(auth_ldap_username_expr, ns_args, &p);
	if (p == NULL) {
	  errmsg = "username string replacement failed";
	  goto fail;
	}
	mapped_username = p;
  }

  role_str = "";
  if ((v = kwv_lookup(kwv, "LDAP_ROLES_SELECTOR*")) != NULL) {
	if (dsvec_len(avas) > 0) {
	  kwv_add(kwv_attrs, "USERNAME", mapped_username);
	  if ((role_str = make_roles(v, kwv_attrs)) != NULL) {
		if (!local_ldap_roles) {
		  ok.roles_reply = ALLOC(Roles_reply);
		  ok.roles_reply->ok = ALLOC(Roles_reply_ok);
		  ok.roles_reply->ok->roles = role_str;
		  ok.roles_reply->failed = NULL;
		  ok.roles_reply->status = NULL;
		}

		log_msg((LOG_TRACE_LEVEL, "Returning role string: \"%s\"", role_str));
	  }
	}
	else
	  log_msg((LOG_TRACE_LEVEL, "No LDAP attributes for roles selection"));
  }
  else
	log_msg((LOG_TRACE_LEVEL, "No LDAP_ROLES_SELECTOR* provided"));

  if (local_ldap_roles)
	emit_roles_reply_ok(stdout, role_str);
  else {
	ok.username = mapped_username;
	printf("%s\n", make_xml_auth_reply_ok(&ok));
	emit_xml_trailer(stdout);
  }

  exit(0);
}
